#
# Cambalache UI wrapper
#
# Copyright (C) 2021  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

import os
from gi.repository import GObject, Gio

from .cmb_path import CmbPath
from .cmb_list_error import CmbListError
from .cmb_base_objects import CmbBaseUI, CmbBaseObject

from cambalache import getLogger, _

logger = getLogger(__name__)


class CmbUI(CmbBaseUI, Gio.ListModel):
    __gtype_name__ = "CmbUI"

    __gsignals__ = {
        "library-changed": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
    }

    path_parent = GObject.Property(type=CmbPath, flags=GObject.ParamFlags.READWRITE)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.update_file_monitor(self.filename)

        self.connect("notify", self.__on_notify)

    def __bool__(self):
        # Override Truth Value Testing to ensure that CmbUI objects evaluates to True even if it does not have children objects
        return True

    def __str__(self):
        return f"CmbUI<{self.display_name}>"

    @GObject.Property(type=int)
    def template_id(self):
        retval = self.db_get("SELECT template_id FROM ui WHERE (ui_id) IS (?);", (self.ui_id,))
        return retval if retval is not None else 0

    @template_id.setter
    def _set_template_id(self, value):
        self.db_set("UPDATE ui SET template_id=? WHERE (ui_id) IS (?);", (self.ui_id,), value if value != 0 else None)

    def __on_notify(self, obj, pspec):
        if pspec.name not in ["file-status", "path-parent"]:
            self.project._ui_changed(self, pspec.name)

        # Update display name if one of the following properties changed
        if pspec.name in ["filename", "template-id", "file-status"]:
            self.notify("display-name")

        if pspec.name == "filename":
            self.update_file_monitor(self.filename)

    def list_libraries(self):
        retval = {}

        for row in self.project.db.execute(
            """
            SELECT DISTINCT t.library_id, NULL
            FROM object AS o, type AS t
            WHERE t.library_id IS NOT NULL AND o.ui_id=? AND o.type_id = t.type_id
            UNION
            SELECT library_id, version FROM ui_library WHERE ui_id=?
            """,
            (self.ui_id, self.ui_id),
        ).fetchall():
            library_id, version = row

            versions = []
            for row in self.project.db.execute(
                "SELECT version FROM library_version WHERE library_id=? ORDER BY version COLLATE version DESC;", (library_id,)
            ).fetchall():
                versions.append(row[0])

            retval[library_id] = {"target": version, "versions": versions}

        return retval

    def get_library(self, library_id):
        c = self.project.db.execute("SELECT version FROM ui_library WHERE ui_id=? AND library_id=?;", (self.ui_id, library_id))
        row = c.fetchone()
        return row[0] if row is not None else None

    def _library_changed(self, lib):
        self.emit("library-changed", lib)
        self.project._ui_library_changed(self, lib)

    def set_library(self, library_id, version, comment=None):
        c = self.project.db.cursor()

        try:
            if version is None:
                c.execute("DELETE FROM ui_library WHERE ui_id=? AND library_id=?;", (self.ui_id, library_id))
            else:
                # Do not use REPLACE INTO, to make sure both INSERT and UPDATE triggers are used
                count = self.db_get(
                    "SELECT count(version) FROM ui_library WHERE ui_id=? AND library_id=?;", (self.ui_id, library_id)
                )

                if count:
                    c.execute(
                        "UPDATE ui_library SET version=?, comment=? WHERE ui_id=? AND library_id=?;",
                        (str(version), comment, self.ui_id, library_id),
                    )
                else:
                    c.execute(
                        "INSERT INTO ui_library (ui_id, library_id, version, comment) VALUES (?, ?, ?, ?);",
                        (self.ui_id, library_id, str(version), comment),
                    )

            self._library_changed(library_id)
        except Exception as e:
            logger.warning(f"{self} Error setting library {library_id}={version}: {e}")

        c.close()

    @classmethod
    def get_display_name(cls, ui_id, filename):
        return os.path.basename(filename) if filename else _("Unnamed {ui_id}").format(ui_id=ui_id)

    @GObject.Property(type=str)
    def display_name(self):
        filename = self.filename
        template_id = self.template_id

        if filename is None and template_id:
            template = self.project.get_object_by_id(self.ui_id, template_id)
            if template:
                return template.name

        display_name = CmbUI.get_display_name(self.ui_id, filename)
        return f'<span underline="error">{display_name}</span>' if self.file_status else display_name

    def __get_infered_target(self, library_id):
        ui_id = self.ui_id

        row = self.project.db.execute(
            """
            WITH lib_version(version) AS (
                SELECT t.version
                FROM object AS o, type AS t
                WHERE t.library_id=? AND o.ui_id=? AND o.type_id = t.type_id AND t.version IS NOT NULL
                UNION
                SELECT p.version
                FROM object_property AS o, property AS p, type AS t
                WHERE t.library_id=? AND o.ui_id=? AND o.owner_id = t.type_id AND o.owner_id = p.owner_id
                  AND p.version IS NOT NULL
                UNION
                SELECT s.version
                FROM object_signal AS o, signal AS s, type AS t
                WHERE t.library_id=? AND o.ui_id=? AND o.owner_id = t.type_id AND o.owner_id = s.owner_id
                  AND s.version IS NOT NULL
            )
            SELECT MAX_VERSION(version) FROM lib_version;
            """,
            (library_id, ui_id, library_id, ui_id, library_id, ui_id),
        ).fetchone()

        return row[0] if row is not None else None

    def get_target(self, library_id):
        target = self.get_library(library_id)
        if target is None:
            target = self.__get_infered_target(library_id)

        if target is None:
            info = self.project.library_info.get(library_id, None)
            if info:
                return info.min_version

        return target

    def reload(self):
        if not self.project or not self.filename:
            return False

        # Disable history
        self.project.history_enabled = False

        # Import file and overwrite
        ui, msgs, detail_msg = self.project.import_file(self.filename, overwrite=True)

        super().reload()

        # Select currently reloaded file
        self.project.set_selection([ui])

    # GListModel iface
    def do_get_item(self, position):
        ui_id = self.ui_id

        # This query should use auto index from UNIQUE constraint
        retval = self.db_get(
            """
            SELECT object_id
            FROM (
                SELECT ROW_NUMBER() OVER (ORDER BY position ASC) rownum, object_id
                FROM object
                WHERE ui_id=? AND parent_id IS NULL
            )
            WHERE rownum=?;
            """,
            (ui_id, position+1)
        )
        if retval is not None:
            return self.project.get_object_by_id(ui_id, retval)

        # This should not happen
        return CmbListError()

    def do_get_item_type(self):
        return CmbBaseObject

    @GObject.Property(type=int)
    def n_items(self):
        retval = self.db_get("SELECT COUNT(object_id) FROM object WHERE ui_id=? AND parent_id IS NULL;", (self.ui_id,))
        return retval if retval is not None else 0

    def do_get_n_items(self):
        return self.n_items




#
# CmbPixbufEntry
#
# Copyright (C) 2021-2023  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from cambalache import _
from gi.repository import Gtk


from .cmb_file_entry import CmbFileEntry


class CmbPixbufEntry(CmbFileEntry):
    __gtype_name__ = "CmbPixbufEntry"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.title = (_("Select Image"),)
        self.props.placeholder_text = "<GdkPixbuf>"

        # Only show images formats supported by GdkPixbuf
        self.filter = Gtk.FileFilter()
        self.filter.add_pixbuf_formats()



#
# Cambalache notification system
#
# Copyright (C) 2025  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

import os
import json
import threading
import http.client
import time
import platform

from uuid import uuid4
from urllib.parse import urlparse
from .config import VERSION
from gi.repository import GObject, GLib, Gio, Gdk, Gtk, Adw, HarfBuzz
from cambalache import getLogger
from . import utils

logger = getLogger(__name__)


class CmbBaseData(GObject.GObject):
    def __init__(self, **kwargs):
        for prop in self.list_properties():
            name = prop.name.replace("-", "_")

            if name not in kwargs:
                continue

            value = kwargs[name]

            if isinstance(value, dict) and prop.value_type in GTYPE_PTYHON:
                Klass = GTYPE_PTYHON[prop.value_type]
                kwargs[name] = Klass(**value)

        super().__init__(**kwargs)

    def dict(self):
        retval = {}

        for prop in self.list_properties():
            name = prop.name.replace("-", "_")

            value = self.get_property(prop.name)

            if prop.value_type in GTYPE_PTYHON:
                retval[name] = value.dict() if value else None
            elif not isinstance(value, GObject.Object):
                retval[name] = value

        return retval


class CmbPollData(CmbBaseData):
    id = GObject.Property(type=str, flags=GObject.ParamFlags.READWRITE)
    title = GObject.Property(type=str, flags=GObject.ParamFlags.READWRITE)
    description = GObject.Property(type=str, flags=GObject.ParamFlags.READWRITE)
    options = GObject.Property(type=object, flags=GObject.ParamFlags.READWRITE)
    allowed_votes = GObject.Property(type=int, default=1, flags=GObject.ParamFlags.READWRITE)
    start_date = GObject.Property(type=int, flags=GObject.ParamFlags.READWRITE)
    end_date = GObject.Property(type=int, flags=GObject.ParamFlags.READWRITE)


class CmbPollResult(CmbBaseData):
    votes = GObject.Property(type=object, flags=GObject.ParamFlags.READWRITE)
    total = GObject.Property(type=int, flags=GObject.ParamFlags.READWRITE)


GTYPE_PTYHON = {CmbPollData.__gtype__: CmbPollData, CmbPollResult.__gtype__: CmbPollResult}


class CmbNotification(CmbBaseData):
    center = GObject.Property(type=GObject.Object, flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT_ONLY)

    type = GObject.Property(type=str, flags=GObject.ParamFlags.READWRITE)
    start_date = GObject.Property(type=int, flags=GObject.ParamFlags.READWRITE)
    end_date = GObject.Property(type=int, flags=GObject.ParamFlags.READWRITE)


class CmbVersionNotification(CmbNotification):
    version = GObject.Property(type=str, flags=GObject.ParamFlags.READWRITE)
    release_notes = GObject.Property(type=str, flags=GObject.ParamFlags.READWRITE)
    read_more_url = GObject.Property(type=str, flags=GObject.ParamFlags.READWRITE)


class CmbMessageNotification(CmbNotification):
    title = GObject.Property(type=str, flags=GObject.ParamFlags.READWRITE)
    message = GObject.Property(type=str, flags=GObject.ParamFlags.READWRITE)


class CmbPollNotification(CmbNotification):
    poll = GObject.Property(type=CmbPollData, flags=GObject.ParamFlags.READWRITE)
    results = GObject.Property(type=CmbPollResult, flags=GObject.ParamFlags.READWRITE)
    my_votes = GObject.Property(type=object, flags=GObject.ParamFlags.READWRITE)


class CmbNotificationCenter(GObject.GObject):
    __gsignals__ = {
        "new-notification": (GObject.SignalFlags.RUN_FIRST, None, (CmbNotification,)),
    }

    # Settings
    enabled = GObject.Property(type=bool, default=True, flags=GObject.ParamFlags.READWRITE)
    uuid = GObject.Property(type=str, flags=GObject.ParamFlags.READWRITE)
    next_request = GObject.Property(type=int, flags=GObject.ParamFlags.READWRITE)
    notifications = GObject.Property(type=str, flags=GObject.ParamFlags.READWRITE)

    store = GObject.Property(type=Gio.ListStore, flags=GObject.ParamFlags.READWRITE)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.retry_interval = 2
        self.user_agent = self.__get_user_agent()
        self.store = Gio.ListStore(item_type=CmbNotification)
        self.settings = Gio.Settings(schema_id="ar.xjuan.Cambalache.notification")

        for prop in ["enabled", "uuid", "next-request", "notifications"]:
            self.settings.bind(prop, self, prop.replace("-", "_"), Gio.SettingsBindFlags.DEFAULT)

        backend = urlparse(os.environ.get("CMB_NOTIFICATION_URL", "https://xjuan.ar:1934"))
        if backend.hostname == "localhost":
            self.REQUEST_INTERVAL = 4
            self.enabled = True
        else:
            self.REQUEST_INTERVAL = 24 * 60 * 60

            # Disable notifications if settings backend is ephemeral
            if GObject.type_name(Gio.SettingsBackend.get_default()) == "GMemorySettingsBackend":
                logger.info("Disabling notifications")
                self.enabled = False

        self.__load_notifications()

        if backend.scheme == "https":
            logger.info(f"Backend: {backend.scheme}://{backend.hostname}:{backend.port}")
            self.connection = http.client.HTTPSConnection(backend.hostname, backend.port, timeout=8)
        else:
            self.connection = None
            logger.warning(f"{backend.scheme} is not supported, only HTTPS")
            return

        # Ensure we have a UUID
        if not self.uuid:
            self.uuid = str(uuid4())

        logger.info(f"User Agent: {self.user_agent}")
        logger.info(f"UUID: {self.uuid}")

        # Wait a few seconds before pooling
        GLib.timeout_add_seconds(4, self._get_notification)

    def __get_container(self):
        if "FLATPAK_ID" in os.environ:
            return "flatpak"
        elif "APPIMAGE" in os.environ:
            return "appimage"
        elif "SNAP" in os.environ:
            return "snap"
        return None

    def __get_user_agent(self):
        u = platform.uname()
        platform_strings = []
        table = str.maketrans({",": "\\,"})

        if u.system == "Linux":
            release = platform.freedesktop_os_release()
            system = f"Linux {release['ID']}"

            if "VERSION_ID" in release:
                system += f" {release['VERSION_ID']}"
            if "VERSION_CODENAME" in release:
                system += f" {release['VERSION_CODENAME']}"
        else:
            system = u.system

        display_type = GObject.type_name(Gdk.Display.get_default())
        backend = display_type.removeprefix("Gdk").removesuffix("Display")
        lang = HarfBuzz.language_to_string(HarfBuzz.language_get_default())
        extra = []

        # Container type
        container = self.__get_container()
        if container:
            extra.append(f"container {container}")

        # GSettings backend
        settings_backend = Gio.SettingsBackend.get_default()
        gsettings_backend = GObject.type_name(settings_backend).removesuffix("SettingsBackend")

        for name, lib in [("GLib", GLib), ("Gtk", Gtk), ("Adw", Adw)]:
            extra.append(f"{name} {lib.MAJOR_VERSION}.{lib.MINOR_VERSION}.{lib.MICRO_VERSION}")

        # Ignore node name as that is private and irrelevant information
        for string in [system, u.release, u.version, u.machine, backend, gsettings_backend]:
            if not string:
                continue
            platform_strings.append(string.translate(table))

        return f"Cambalache/{VERSION} ({', '.join(platform_strings)}; {'; '.join(extra)}; {lang})"

    def __load_notifications(self):
        self.store.remove_all()

        if not self.notifications:
            return

        notifications = json.loads(self.notifications)
        now = utils.utcnow()

        for data in notifications:
            if "end_date" in data and now > data["end_date"]:
                continue

            self.store.append(self.__notification_from_dict(data))

    def __save_notifications(self):
        notifications = []

        for notification in self.store:
            notifications.append(notification.dict())

        # Store in GSettings
        self.notifications = json.dumps(notifications, indent=2, sort_keys=True)

    def __notification_from_dict(self, data):
        ntype = data.get("type", None)

        if ntype == "version":
            return CmbVersionNotification(center=self, **data)
        elif ntype == "message":
            return CmbMessageNotification(center=self, **data)
        elif ntype == "poll":
            return CmbPollNotification(center=self, **data)

    def __get_notification_idle(self, data):
        logger.debug(f"Got notification response {json.dumps(data, indent=2, sort_keys=True)}")

        if "notification" in data:
            notification = self.__notification_from_dict(data["notification"])
            self.store.insert(0, notification)
            self.__save_notifications()
            self.emit("new-notification", notification)

        now = int(time.time())
        self.next_request = now + self.REQUEST_INTERVAL
        self._get_notification()

        return GLib.SOURCE_REMOVE

    def __get_notification_thread(self):
        headers = {
            "User-Agent": self.user_agent,
            "x-cambalache-uuid": self.uuid,
        }

        try:
            logger.info(f"GET /notification {headers=}")

            self.connection.request("GET", "/notification", headers=headers)
            response = self.connection.getresponse()
            assert response.status == 200

            # Reset retry interval
            self.retry_interval = 8

            data = response.read().decode()

            logger.info(f"response={data}")

            if data:
                GLib.idle_add(self.__get_notification_idle, json.loads(data))
        except Exception as e:
            # If it fails we just wait a bit before retrying
            self.retry_interval *= 2
            self.retry_interval = min(self.retry_interval, 256)

            logger.info(f"Request error {e}, retrying in {self.retry_interval}s")
            GLib.timeout_add_seconds(self.retry_interval, self._get_notification)

        self.connection.close()

    def __run_in_thread(self, function, *args, **kwargs):
        if not self.connection:
            logger.warning("No connection defined")
            return

        if not self.enabled:
            logger.info("Notifications disabled")
            return

        thread = threading.Thread(target=function, args=args, kwargs=kwargs, daemon=True)
        thread.start()

    def _get_notification(self):
        now = int(time.time())

        if now >= self.next_request:
            self.__run_in_thread(self.__get_notification_thread)
        else:
            GLib.timeout_add_seconds(self.next_request - now, self._get_notification)

    def __poll_vote_idle(self, data):
        logger.debug(f"Got vote response {data}")

        poll_uuid = data["uuid"]
        results = data["results"]

        for notification in self.store:
            if isinstance(notification, CmbPollNotification) and notification.poll.id == poll_uuid:
                notification.results = CmbPollResult(**results)
                self.__save_notifications()
                break
        return GLib.SOURCE_REMOVE

    def __poll_vote_exception_idle(self, poll_uuid):
        for notification in self.store:
            if isinstance(notification, CmbPollNotification) and notification.poll.id == poll_uuid:
                notification.my_votes = []
                break
        return GLib.SOURCE_REMOVE

    def __poll_vote_thread(self, method, poll_uuid, votes=None):
        headers = {"User-Agent": self.user_agent, "x-cambalache-uuid": self.uuid, "Content-type": "application/json"}

        try:
            payload = json.dumps({"votes": votes}) if method == "POST" else None
            self.connection.request(method, f"/poll/{poll_uuid}", payload, headers)
            response = self.connection.getresponse()
            assert response.status == 200

            data = response.read().decode()
            GLib.idle_add(self.__poll_vote_idle, json.loads(data))
        except Exception as e:
            logger.warning(f"Error voting {e}")
            GLib.idle_add(self.__poll_vote_exception_idle, poll_uuid)

        self.connection.close()

    def poll_vote(self, notification: CmbPollNotification, votes: list[int]):
        if self.uuid is None:
            return

        notification.my_votes = votes

        self.__run_in_thread(self.__poll_vote_thread, "POST", notification.poll.id, votes)

    def poll_refresh_results(self, notification: CmbPollNotification):
        if self.uuid is None:
            return

        self.__run_in_thread(self.__poll_vote_thread, "GET", notification.poll.id)

    def remove(self, notification: CmbNotification):
        valid, position = self.store.find(notification)
        if valid:
            self.store.remove(position)
            self.__save_notifications()

    def remove_all(self):
        self.store.remove_all()
        self.__save_notifications()


notification_center = CmbNotificationCenter()



#
# CmbTypeChooserPopover - Cambalache Type Chooser Popover
#
# Copyright (C) 2021-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk

from .cmb_project import CmbProject
from .cmb_type_info import CmbTypeInfo
from .cmb_type_chooser_widget import CmbTypeChooserWidget


class CmbTypeChooserPopover(Gtk.Popover):
    __gtype_name__ = "CmbTypeChooserPopover"

    __gsignals__ = {
        "type-selected": (GObject.SignalFlags.RUN_LAST, None, (CmbTypeInfo,)),
    }

    project = GObject.Property(type=CmbProject, flags=GObject.ParamFlags.READWRITE)
    category = GObject.Property(type=str, flags=GObject.ParamFlags.READWRITE)
    uncategorized_only = GObject.Property(type=bool, flags=GObject.ParamFlags.READWRITE, default=False)
    show_categories = GObject.Property(type=bool, flags=GObject.ParamFlags.READWRITE, default=False)
    parent_type_id = GObject.Property(type=str, flags=GObject.ParamFlags.READWRITE)
    derived_type_id = GObject.Property(type=str, flags=GObject.ParamFlags.READWRITE)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._chooser = CmbTypeChooserWidget()
        self._chooser.connect("type-selected", self.__on_type_selected)

        self.set_child(self._chooser)
        self.set_default_widget(self._chooser)

        for prop in [
            "project",
            "category",
            "uncategorized_only",
            "show_categories",
            "parent_type_id",
            "derived_type_id",
        ]:
            GObject.Object.bind_property(self, prop, self._chooser, prop, GObject.BindingFlags.SYNC_CREATE)

    def __on_type_selected(self, chooser, info):
        self.emit("type-selected", info)
        self.popdown()



# GBinding Workspace proxy
#
# Copyright (C) 2022  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject


#
# PROBLEM: GBinding calls g_assert() in its constructor to check all construct properties
# are set correctly if not, it terminates the current process.
#
# WORKAROUND: only create the GBinding once we have all the properties
#
class MrgGBindingProxy(GObject.Object):
    __gtype_name__ = "MrgGBindingProxy"

    flags = GObject.Property(
        type=GObject.BindingFlags, default=GObject.BindingFlags.DEFAULT, flags=GObject.ParamFlags.READWRITE
    )
    source = GObject.Property(type=GObject.Object, flags=GObject.ParamFlags.READWRITE)
    source_property = GObject.Property(type=str, flags=GObject.ParamFlags.READWRITE)
    target = GObject.Property(type=GObject.Object, flags=GObject.ParamFlags.READWRITE)
    target_property = GObject.Property(type=str, flags=GObject.ParamFlags.READWRITE)

    def __init__(self, **kwargs):
        self.__proxy = None

        super().__init__(**kwargs)

        self.connect("notify", self.__on_notify)

    def __on_notify(self, obj, pspec):
        if self.__proxy:
            self.__proxy.unbind()
            self.__proxy = None

        if (
            self.flags
            and self.source
            and self.source_property
            and self.target
            and self.target_property
            and self.source.find_property(self.source_property)
            and self.target.find_property(self.target_property)
        ):
            self.__proxy = self.source.bind_property(self.source_property, self.target, self.target_property, self.flags)



#
# CmbTreeExpander
#
# Copyright (C) 2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gdk, Gtk, Graphene

from .cmb_ui import CmbUI
from .cmb_object import CmbObject
from .cmb_css import CmbCSS
from .cmb_path import CmbPath
from cambalache import _


class CmbTreeExpander(Gtk.TreeExpander):
    __gtype_name__ = "CmbTreeExpander"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.__project = None
        self.__drop_before = None

        self.label = Gtk.Inscription(hexpand=True)
        self.set_child(self.label)

        self.__parent = None
        self.__drag_source = None
        self.__drop_target = None

    def update_bind(self):
        item = self.props.item

        self.__parent = self.props.parent
        # Drag source
        self.__drag_source = Gtk.DragSource()
        self.__drag_source.connect("prepare", self.__on_drag_prepare)
        self.__drag_source.connect("drag-begin", self.__on_drag_begin)
        self.__parent.add_controller(self.__drag_source)

        # Drop target
        self.__drop_target = Gtk.DropTarget.new(type=GObject.TYPE_NONE, actions=Gdk.DragAction.COPY)
        self.__drop_target.set_gtypes([CmbObject, CmbUI])
        self.__drop_target.connect("accept", self.__on_drop_accept)
        self.__drop_target.connect("motion", self.__on_drop_motion)
        self.__drop_target.connect("drop", self.__on_drop_drop)
        self.__parent.add_controller(self.__drop_target)

        # Handle label
        self.__update_label(item)
        item.connect("notify::display-name", self.__on_item_display_name_notify)

        self.__project = item.project

        if isinstance(item, CmbCSS):
            self.props.hide_expander = True
            return
        elif isinstance(item, CmbPath):
            self.props.hide_expander = False
            self.add_css_class("cmb-path" if item.path else "cmb-unsaved-path")
            return

        self.props.hide_expander = item.props.n_items == 0
        item.connect("notify::n-items", self.__on_item_n_items_notify)

    def clear_bind(self):
        item = self.props.item

        if self.__parent:
            self.__parent.remove_controller(self.__drag_source)
            self.__parent.remove_controller(self.__drop_target)
            self.__parent = None
            self.__drag_source = None
            self.__drop_target = None

        self.remove_css_class("cmb-path")
        self.remove_css_class("cmb-unsaved-path")

        item.disconnect_by_func(self.__on_item_display_name_notify)

        if isinstance(item, CmbCSS) or isinstance(item, CmbPath):
            return

        item.disconnect_by_func(self.__on_item_n_items_notify)

    def __on_item_n_items_notify(self, item, pspec):
        self.props.hide_expander = item.props.n_items == 0

    def __on_item_display_name_notify(self, item, pspec):
        self.__update_label(item)

    def __update_label(self, item):
        self.label.set_markup(item.props.display_name or "")

    # Drop target callbacks
    def __get_drop_before(self, widget, x, y):
        h = widget.get_height()

        if y < h/4:
            return True

        if y > h * 0.8:
            return False

        return None

    def __on_object_drop_accept(self, drop, item):
        origin_item = drop.get_drag()._item

        if origin_item == item:
            return None

        if not isinstance(item, CmbObject):
            return None

        # Ignore if its the same parent
        if origin_item.parent_id == item.object_id:
            return None

        return origin_item

    def __on_drop_accept(self, target, drop):
        item = self.props.item
        origin_item = drop.get_drag()._item

        if isinstance(item, CmbObject):
            origin_item = self.__on_object_drop_accept(drop, item)

            self.__drop_before = None

            if origin_item is None or item.parent is None:
                return False

            return self.__project._check_can_add(origin_item.type_id, item.parent.type_id)
        elif isinstance(item, CmbUI):
            if origin_item == item:
                return False

            # Ignore if its the same UI and item is already a toplevel
            if origin_item.ui_id == item.ui_id and origin_item.parent_id is None:
                return False

            return True

    def __on_drop_motion(self, target, x, y):
        item = self.props.item

        if isinstance(item, CmbObject):
            row_widget = target.get_widget()

            drop_before = self.__get_drop_before(row_widget, x, y)

            if self.__drop_before == drop_before:
                return Gdk.DragAction.COPY

            row_widget.remove_css_class("drop-before")
            row_widget.remove_css_class("drop-after")

            self.__drop_before = drop_before

            if drop_before is not None:
                row_widget.add_css_class("drop-before" if drop_before else "drop-after")

        return Gdk.DragAction.COPY

    def __on_drop_drop(self, target, origin_item, x, y):
        row_widget = target.get_widget()
        item = self.props.item

        if isinstance(item, CmbObject):
            drop_before = self.__get_drop_before(row_widget, x, y)

            if drop_before is None:
                # TODO: handle dragging from one UI to another
                if origin_item.ui_id != item.ui_id:
                    return

                if origin_item.parent_id != item.object_id:
                    self.__project.history_push(
                        _("Move {name} to {target}").format(name=origin_item.display_name, target=item.display_name)
                    )
                    origin_item.parent_id = item.object_id
                    self.__project.history_pop()

                return

            if origin_item.parent_id != item.parent_id:
                if drop_before:
                    msg = _("Move {name} before {target}").format(name=origin_item.display_name, target=item.display_name)
                else:
                    msg = _("Move {name} after {target}").format(name=origin_item.display_name, target=item.display_name)

                self.__project.history_push(msg)

                origin_item.parent_id = item.parent_id
            else:
                msg = None

            parent = item.parent

            origin_position = origin_item.position
            target_position = item.position

            if origin_position > target_position:
                if drop_before:
                    position = item.position
                else:
                    position = item.position + 1
            else:
                if drop_before:
                    position = item.position - 1
                else:
                    position = item.position

            parent.reorder_child(origin_item, position)

            if msg:
                self.__project.history_pop()

            self.__project.set_selection([origin_item])
        elif isinstance(item, CmbUI):
            if origin_item.ui_id == item.ui_id:
                self.__project.history_push(_("Move {name} as toplevel").format(name=origin_item.display_name))
                origin_item.parent_id = 0
                self.__project.history_pop()
            else:
                # TODO: Use copy/paste to move across UI files
                pass

    # Drag Source callbacks
    def __on_drag_prepare(self, drag_source, x, y):
        item = self.props.item

        # Only CmbObject start a drag
        if not isinstance(item, CmbObject):
            return None

        self.__drag_point = Graphene.Point()
        self.__drag_point.x = x
        self.__drag_point.y = y
        return Gdk.ContentProvider.new_for_value(self.props.item)

    def __on_drag_begin(self, drag_source, drag):
        drag._item = self.props.item
        valid, p = self.__parent.compute_point(self.label, self.__drag_point)
        if valid:
            drag_source.set_icon(Gtk.WidgetPaintable.new(self.label), p.x, p.y)



#
# CmbContextMenu - Cambalache UI Editor
#
# Copyright (C) 2021-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

import os

from gi.repository import GObject, GLib, Gio, Gdk, Gtk
from cambalache import _


@Gtk.Template(resource_path="/ar/xjuan/Cambalache/cmb_context_menu.ui")
class CmbContextMenu(Gtk.PopoverMenu):
    __gtype_name__ = "CmbContextMenu"

    enable_theme = GObject.Property(
        type=bool,
        default=False,
        flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT_ONLY
    )
    target_tk = GObject.Property(type=str, flags=GObject.ParamFlags.READWRITE)

    main_section = Gtk.Template.Child()
    add_submenu = Gtk.Template.Child()

    def __init__(self, **kwargs):
        self.theme_submenu = None

        super().__init__(**kwargs)

        self.connect("notify::target-tk", self.__on_target_tk_notify)

    def __on_target_tk_notify(self, obj, pspec):
        self.__populate_css_theme_box()
        self.__update_add_submenu()

    def __update_add_submenu(self):
        if self.target_tk not in ["gtk-4.0", "gtk+-3.0"]:
            return

        types = [
            "GtkBox",
            "GtkGrid",
            "GtkExpander",
            "GtkRevealer",
            "GtkOverlay",
            "GtkScrolledWindow",
        ]

        if self.target_tk == "gtk+-3.0":
            types += [
                "GtkAligment",
                "GtkEventBox"
            ]
        else:
            types += [
                "GtkGraphicsOffload",
            ]

        self.add_submenu.remove_all()

        for gtype in sorted(types):
            item = Gio.MenuItem()
            item.set_label(gtype)
            item.set_action_and_target_value("win.add_parent", GLib.Variant("s", gtype))
            self.add_submenu.append_item(item)

    def __populate_css_theme_box(self):
        gtk_path = "gtk-3.0"

        if not self.enable_theme or self.target_tk not in ["gtk-4.0", "gtk+-3.0"]:
            return

        if self.target_tk == "gtk-4.0":
            gtk_path = "gtk-4.0"
            # FIXME: whats the real default theme for gtk4?
            themes = ["Default"]
        else:
            themes = ["Adwaita", "HighContrast", "HighContrastInverse"]

        if self.theme_submenu is None:
            self.theme_submenu = Gio.Menu()
            self.main_section.prepend_submenu(_("CSS theme"), self.theme_submenu)

        # Remove all items from theme submenu
        self.theme_submenu.remove_all()

        dirs = []

        dirs += GLib.get_system_data_dirs()
        dirs.append(GLib.get_user_data_dir())

        # Add /themes to every dir
        dirs = list(map(lambda d: os.path.join(d, "themes"), dirs))

        # Append ~/.themes
        dirs.append(os.path.join(GLib.get_home_dir(), ".themes"))

        for path in dirs:
            if not os.path.isdir(path):
                continue

            for theme in os.listdir(path):
                tpath = os.path.join(path, theme, gtk_path, "gtk.css")
                if os.path.exists(tpath):
                    themes.append(theme)

        # Dedup and sort
        themes = list(dict.fromkeys(themes))

        for theme in sorted(themes):
            item = Gio.MenuItem()
            item.set_label(theme)
            item.set_action_and_target_value("win.workspace_theme", GLib.Variant("s", theme))
            self.theme_submenu.append_item(item)

    def popup_at(self, x, y):
        r = Gdk.Rectangle()
        r.x, r.y = (x, y)
        r.width = r.height = 0
        self.set_pointing_to(r)
        self.popup()



#
# CmbObjectPropertyEditor - Cambalache Object Editor
#
# Copyright (C) 2021-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk

from .cmb_object import CmbObject
from .cmb_object_data_editor import CmbObjectDataEditor
from .control import CmbEntry, CmbChildTypeComboBox, CmbSourceView, cmb_create_editor
from .cmb_property_label import CmbPropertyLabel
from cambalache import _
from .constants import EXTERNAL_TYPE, CUSTOM_TYPE
from . import utils


class CmbObjectPropertyEditor(Gtk.Box):
    __gtype_name__ = "CmbObjectPropertyEditor"

    layout = GObject.Property(type=bool, flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT_ONLY, default=False)

    def __init__(self, **kwargs):
        self.__object = None
        self.__id_label = None
        self.__type_label = None
        self.__template_switch = None
        self.__bindings = []

        super().__init__(**kwargs)

        self.props.orientation = Gtk.Orientation.VERTICAL

    def bind_property(self, s, sp, t, tp, flags=GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL):
        binding = GObject.Object.bind_property(s, sp, t, tp, flags)
        self.__bindings.append(binding)
        return binding

    def __create_id_editor(self):
        grid = Gtk.Grid(hexpand=True, row_spacing=4, column_spacing=4)

        # Label
        self.__id_label = Gtk.Label(label=_("Object Id"), halign=Gtk.Align.START)

        # Id/Class entry
        entry = CmbEntry()
        self.bind_property(self.__object, "name", entry, "cmb-value")

        i = 0
        grid.attach(self.__id_label, 0, i, 1, 1)
        grid.attach(entry, 1, i, 1, 1)
        i += 1

        if self.__object.type_id == CUSTOM_TYPE:
            editor = CmbEntry(halign=Gtk.Align.START)

            prop = self.__object.properties_dict.get("type")
            self.bind_property(prop, "value", editor, "cmb-value")

            self.__type_label = Gtk.Label(label=_("Type"), halign=Gtk.Align.START)

            grid.attach(self.__type_label, 0, i, 1, 1)
            grid.attach(editor, 1, i, 1, 1)
            i += 1

        # Template check
        if self.__object and self.__object.parent_id == 0:
            is_template = self.__object.object_id == self.__object.ui.template_id
            tooltip_text = _("Switch between object and template")
            derivable = self.__object.info.derivable

            if not derivable:
                tooltip_text = _("{type} is not derivable.").format(type=self.__object.info.type_id)

            label = Gtk.Label(label=_("Template"), halign=Gtk.Align.START, tooltip_text=tooltip_text, sensitive=derivable)
            self.__template_switch = Gtk.Switch(
                active=is_template, halign=Gtk.Align.START, tooltip_text=tooltip_text, sensitive=derivable
            )

            self.__template_switch.connect("notify::active", self.__on_template_switch_notify)
            self.__update_template_label()

            grid.attach(label, 0, i, 1, 1)
            grid.attach(self.__template_switch, 1, i, 1, 1)

        return grid

    def __on_shortcut_button_clicked(self, button, type_id):
        obj = self.__object
        obj.project.add_object(obj.ui_id, type_id, parent_id=obj.object_id)

    def __create_child_shortcuts(self, info):
        box = Gtk.FlowBox(visible=True, hexpand=True, selection_mode=Gtk.SelectionMode.NONE)

        label = Gtk.Label(label=_("Add"), xalign=0, visible=True)
        box.append(label)

        for type_id in info.child_type_shortcuts:
            button = Gtk.Button(label=type_id, visible=True)
            button.connect("clicked", self.__on_shortcut_button_clicked, type_id)
            box.append(button)

        return box

    def __update_template_label(self):
        istmpl = self.__object.ui.template_id == self.__object.object_id
        self.__id_label.props.label = _("Type Name") if istmpl else _("Object Id")

        if self.__type_label:
            self.__type_label.props.label = _("Parent Class") if istmpl else _("Type")

    def __on_template_switch_notify(self, switch, pspec):
        self.__object.ui.template_id = self.__object.object_id if switch.props.active else 0
        self.__update_template_label()

    def __on_expander_expanded(self, expander, pspec, revealer):
        expanded = expander.props.expanded

        if expanded:
            revealer.props.transition_type = Gtk.RevealerTransitionType.SLIDE_DOWN
        else:
            revealer.props.transition_type = Gtk.RevealerTransitionType.SLIDE_UP

        revealer.props.reveal_child = expanded

    def __create_child_type_editor(self):
        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)

        box.append(Gtk.Label(label=_("Child Type"), width_chars=8))

        combo = CmbChildTypeComboBox(object=self.__object)

        self.bind_property(self.__object, "type", combo, "cmb-value")
        box.append(combo)
        return box

    def __update_view(self):
        for child in utils.widget_get_children(self):
            self.remove(child)

        if self.__object is None:
            return

        obj = self.__object
        parent = obj.parent
        is_builtin = obj.info.is_builtin

        if self.layout:
            if parent is None or is_builtin:
                return

            # Child Type input
            if parent.info.has_child_types():
                self.append(self.__create_child_type_editor())
        elif not obj.info.is_a("GtkExpression"):
            # ID
            self.append(self.__create_id_editor())

        if obj.type_id == EXTERNAL_TYPE:
            label = Gtk.Label(
                label=_(
                    "This object will not be exported, it is only used to make a reference to it. \
It has to be exposed by your application with GtkBuilder expose_object method."
                ),
                halign=Gtk.Align.START,
                margin_top=8,
                xalign=0,
                wrap=True,
            )
            self.append(label)
            self.show()
            return
        elif obj.type_id == CUSTOM_TYPE:
            # Fragment Editor
            sw = Gtk.ScrolledWindow(has_frame=True, vexpand=True, margin_top=6)
            editor = CmbSourceView(lang="xml")
            self.bind_property(obj, "custom-fragment", editor, "text")
            sw.set_child(editor)
            self.append(sw)
            return

        info = parent.info if self.layout and parent else obj.info
        for owner_id in [info.type_id] + info.hierarchy:
            if self.layout:
                owner_id = f"{owner_id}LayoutChild"

            info = obj.project.type_info.get(owner_id, None)

            if info is None:
                continue

            # Editor count
            i = 0

            # Grid for all properties and custom data editors
            grid = Gtk.Grid(hexpand=True, row_spacing=4, column_spacing=4)

            # Add shortcuts
            if not self.layout and len(info.child_type_shortcuts):
                shortcuts = self.__create_child_shortcuts(info)
                shortcuts.props.margin_start = 14
                grid.attach(shortcuts, 0, i, 2, 1)
                i += 1

            # Properties
            properties = obj.layout_dict if self.layout else obj.properties_dict
            for property_id in info.properties:
                prop = properties.get(property_id, None)

                if prop is None or prop.info is None:
                    continue

                # Only show properties in the class they where originally defined
                if prop.info.original_owner_id is not None and owner_id != prop.info.original_owner_id:
                    continue

                editor = cmb_create_editor(prop.project, prop.info.type_id, prop=prop)

                if editor is None:
                    continue

                self.bind_property(prop, "value", editor, "cmb-value")

                if self.layout:
                    label = CmbPropertyLabel(layout_prop=prop)
                else:
                    label = CmbPropertyLabel(prop=prop, bindable=not is_builtin)

                if prop.info.disabled:
                    for w in [editor, label]:
                        w.set_tooltip_text(_("This property is disabled for {type_id}").format(type_id=obj.type_id))
                        w.props.sensitive = False

                grid.attach(label, 0, i, 1, 1)
                grid.attach(editor, 1, i, 1, 1)
                i += 1

            for data_key in info.data:
                data = None

                # Find data
                for d in obj.data:
                    if d.info.key == data_key:
                        data = d
                        break

                editor = CmbObjectDataEditor(
                    visible=True,
                    hexpand=True,
                    object=obj,
                    data=data,
                    info=info.data[data_key],
                )

                grid.attach(editor, 0, i, 2, 1)
                i += 1

            # Continue if class had no editors to add
            if i == 0:
                continue

            # Create expander/revealer to pack editor grid
            expander = Gtk.Expander(label=f"<b>{owner_id}</b>", use_markup=True, expanded=True)
            revealer = Gtk.Revealer(reveal_child=True)
            expander.connect("notify::expanded", self.__on_expander_expanded, revealer)
            revealer.set_child(grid)
            self.append(expander)
            self.append(revealer)

        self.show()

    def __on_object_ui_notify(self, obj, pspec):
        if pspec.name == "template-id" and self.__template_switch:
            self.__template_switch.set_active(obj.props.template_id != 0)

    def __on_object_notify(self, obj, pspec):
        if pspec.name == "parent-id":
            self.__update_view()

    @GObject.Property(type=CmbObject)
    def object(self):
        return self.__object

    @object.setter
    def _set_object(self, obj):
        if obj == self.__object:
            return

        if self.__object:
            self.__object.disconnect_by_func(self.__on_object_notify)
            if self.__object.ui:
                self.__object.ui.disconnect_by_func(self.__on_object_ui_notify)

        for binding in self.__bindings:
            binding.unbind()

        self.__bindings = []

        self.__object = obj

        if obj:
            obj.connect("notify", self.__on_object_notify)
            if obj.ui:
                obj.ui.connect("notify", self.__on_object_ui_notify)

        self.__update_view()


Gtk.WidgetClass.set_css_name(CmbObjectPropertyEditor, "CmbObjectPropertyEditor")



#
# CmbEntry
#
# Copyright (C) 2021-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk


class CmbEntry(Gtk.Entry):
    __gtype_name__ = "CmbEntry"

    __gsignals__ = {
        "edit-translatable": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.connect("notify::text", self.__on_text_notify)
        self.connect("icon-press", self.__on_icon_pressed)

    def __on_icon_pressed(self, widget, icon_pos):
        self.emit("edit-translatable")

    def __on_text_notify(self, obj, pspec):
        self.notify("cmb-value")

    @GObject.Property(type=bool, default=False)
    def translatable(self):
        return self.props.secondary_icon_name == "document-edit-symbolic"

    @translatable.setter
    def _set_translatable(self, value):
        self.props.secondary_icon_name = "document-edit-symbolic" if value else None

    @GObject.Property(type=str)
    def cmb_value(self):
        return self.props.text if self.props.text != "" else None

    @cmb_value.setter
    def _set_cmb_value(self, value):
        # We do not want to emit a change if there is none
        if value == self.props.text:
            return

        self.props.text = value if value is not None else ""



# GtkDialog Controller
#
# Copyright (C) 2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GLib, GObject, Gtk

from .mrg_gtk_window import MrgGtkWindow
from merengue import getLogger

logger = getLogger(__name__)


class MrgGtkDialog(MrgGtkWindow):
    object = GObject.Property(type=Gtk.Dialog, flags=GObject.ParamFlags.READWRITE)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __queue_resize(self, data):
        if self.object:
            self.object.props.border_width = data
        return GLib.SOURCE_REMOVE

    def object_changed(self, old, new):
        super().object_changed(old, new)

        if Gtk.MAJOR_VERSION == 3 and new:
            # FIXME: Hack, force dialog to resize properly
            # GtkDialog gets allocated too much space or not enough for messages, setting message or container border fix it.
            # Probably a bug in Casilda or a workaround in gnome-shell
            GLib.timeout_add(100, self.__queue_resize, new.props.border_width)
            new.props.border_width = 1



#
# CmbGirData - Gir helper
#
# Copyright (C) 2020-2024  Juan Pablo Ugarte
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#

import gi
import importlib
import logging

# We need to use lxml to get access to nsmap
from lxml import etree
from graphlib import TopologicalSorter, CycleError
from gi.repository import GLib, GObject

CmbCatalogUtils = None

# Global XML name space
nsmap = {}


# Helper function to get a namespaced attribute
def ns(namespace, name):
    return f"{{{nsmap[namespace]}}}{name}"


class CmbGirData:
    def __init__(
        self,
        gir_file,
        gtk_version=None,
        libname=None,
        types=None,
        flag_types=None,
        enum_types=None,
        boxed_types=[],
        skip_types=[],
        exclude_objects=False,
        external_types=None,
    ):
        self._instances = {}

        # Supported param specs
        self.pspec_map = {
            "GParamBoolean": "gboolean",
            "GParamChar": "gchar",
            "GParamUChar": "guchar",
            "GParamInt": "gint",
            "GParamUInt": "guint",
            "GParamLong": "glong",
            "GParamULong": "gulong",
            "GParamInt64": "gint64",
            "GParamUInt64": "guint64",
            "GParamFloat": "gfloat",
            "GParamDouble": "gdouble",
            "GParamEnum": "enum",
            "GParamFlags": "flags",
            "GParamString": "gchararray",
            "GParamObject": "object",
            "GParamUnichar": "gunichar",
            "GParamGType": "gtype",
            "GParamBoxed": "boxed",
            "GParamVariant": "variant",
            "GtkParamSpecExpression": "object",
        }

        # GtkBuilder native object types
        self.builder_native_object_types = ["GdkPixbuf", "GFile", "GtkShortcutTrigger", "GtkShortcutAction"]

        tree = etree.parse(gir_file)
        root = tree.getroot()

        # Set global NS map
        global nsmap
        nsmap = root.nsmap

        self.lib = libname

        # Get <namespace/>
        namespace = root.find("namespace", nsmap)

        # Get module/name space data
        self.name = namespace.get("name")

        # Use first prefix by default, add new option if needed.
        self.prefix = namespace.get(ns("c", "identifier-prefixes")).split(",")[0]
        self.version = namespace.get("version")
        self.shared_library = namespace.get("shared-library")
        self.accessibility_metadata = {}

        self.external_types = external_types if external_types else {}

        self.external_nstypes = {}
        for t in self.external_types:
            self.external_nstypes[self.external_types[t]] = t

        self.ignored_pspecs = set()
        self.ignored_types = set()
        self.ignored_boxed_types = set()

        # Load Module described by gir
        try:
            print(f"Loading {self.name} {self.version} with prefix {self.prefix}")
            gi.require_version(self.name, self.version)
            self.mod = importlib.import_module(f"gi.repository.{self.name}")

            if hasattr(self.mod, "init"):
                init_function = getattr(self.mod, "init")
                if len(init_function.get_arguments()) == 0:
                    init_function()

        except ValueError as e:
            logging.error(f"Could not load {self.name} {self.version} module: {e}")

        gi.require_version("CmbCatalogUtils", "4.0" if gtk_version == 4 else "3.0")

        global CmbCatalogUtils
        from gi.repository import CmbCatalogUtils

        # Dictionary of all enumerations
        self.enumerations = self._get_enumerations(namespace, enum_types)

        # Dictionary of all flags
        self.flags = self._get_flags(namespace, flag_types)

        # Include Boxed types
        self.types = self._get_boxed_types(boxed_types)

        # Dictionary of all interfaces
        self.ifaces = self._get_ifaces(namespace, types, exclude_objects)

        # Dictionary of all classes/types
        obj_types = self._get_types(namespace, types, skip_types, exclude_objects)
        self.types.update(obj_types)

        self._cmb_types_init()

        if gtk_version == 4:
            self._gtk4_init()
        elif gtk_version == 3:
            self._gtk3_init()

        # Types dependency graph
        types_deps = {}
        for gtype in self.types:
            dep = self.types[gtype]["parent"]
            if dep:
                types_deps[gtype] = {dep}

        # Types in topological order, to avoid FK errors
        try:
            ts = TopologicalSorter(types_deps)
            self.sorted_types = tuple(ts.static_order())
        except CycleError as e:
            raise Exception(f"WEIRD! Found dependency cycle sorting types {e}")

    def _type_is_a(self, type, is_a_type):
        if type == is_a_type:
            return True

        parent = self.types.get(type, None)

        while parent:
            if parent["parent"] == is_a_type:
                return True
            else:
                parent = self.types.get(parent["parent"], None)

        return False

    def _get_instance_from_type(self, name):
        retval = self._instances.get(name, None)

        if retval is not None:
            return retval

        if name.startswith(self.prefix):
            InstanceClass = getattr(self.mod, name[len(self.prefix) :], None)
        else:
            InstanceClass = getattr(self.mod, name, None)

        required_params = {}

        # TODO: add this to the catalog directly
        if name == "AdwLayoutSlot":
            required_params = {
                "id": "an slot id"
            }
        elif name == "AdwSwipeTracker":
            from gi.repository import Adw
            required_params = {
                "swipeable": Adw.Flap(),
            }
        elif name == "AdwPropertyAnimationTarget":
            class FooBar(GObject.GObject):
                foobar = GObject.Property(type=int, flags=GObject.ParamFlags.READWRITE)

            required_params = {
                "object": FooBar(),
                "pspec": FooBar.find_property("foobar")
            }

        gtype = GObject.type_from_name(name)
        if InstanceClass is not None and GObject.type_is_a(gtype, GObject.GObject):
            if GObject.type_test_flags(gtype, GObject.TypeFlags.ABSTRACT):
                # Ensure class is instantiable
                class ChildClass(InstanceClass):
                    pass

                if ChildClass is not None:
                    retval = ChildClass(**required_params)
            else:
                retval = InstanceClass(**required_params)

        # keep the instance for later
        if retval is not None:
            self._instances[name] = retval

        return retval

    def _container_list_child_properties(self, name):
        instance = self._get_instance_from_type(name)

        if instance is not None:
            props = instance.list_child_properties()
            return props if len(props) > 0 else None

        return None

    def _type_get_properties_overrides(self, name):
        if name == "GObject":
            return []

        class_type = GObject.type_from_name(name)
        parent_type = GObject.type_parent(class_type)
        parent_name = GObject.type_name(parent_type)

        class_interfaces = GObject.type_interfaces(class_type)
        parent_interfaces = GObject.type_interfaces(parent_type)

        instance = self._get_instance_from_type(name)
        if instance is None:
            return []

        parent_instance = self._get_instance_from_type(parent_name)
        retval = []

        for pspec in instance.list_properties():
            writable = pspec.flags & GObject.ParamFlags.WRITABLE
            readable = pspec.flags & GObject.ParamFlags.READABLE
            owner = pspec.owner_type

            # Ignore properties that can not be written or read
            if not writable or not readable:
                continue

            # Object, Boxed and GType params do not have a default value
            if GObject.type_is_a(pspec.value_type, GObject.TYPE_BOXED):
                continue

            if GObject.type_is_a(pspec.value_type, GObject.TYPE_GTYPE):
                continue

            value_type = GObject.type_name(pspec.value_type)
            if GObject.type_is_a(pspec.value_type, GObject.TYPE_OBJECT) and value_type not in self.types:
                continue

            instance_default = getattr(instance.props, pspec.name)
            if pspec.get_default_value() == instance_default:
                continue

            if owner == class_type or (owner in class_interfaces and owner not in parent_interfaces):
                # This is a property declared in this class
                # We need to make sure the default declared in the pspec is the same as the instance
                retval.append(
                    {
                        "property_id": pspec.name,
                        "instance_default":
                            self._get_default_value_from_pspec(pspec, instance_default, use_instance_default=True)
                    }
                )
            elif parent_instance:
                parent_instance_default = getattr(parent_instance.props, pspec.name)

                # Ignore GObject properties for now
                # if a subclass sets a parent object property that probably means we should disable it because its being used.
                if GObject.type_is_a(pspec.value_type, GObject.TYPE_OBJECT):
                    if parent_instance_default is None and instance_default is not None:
                        retval.append({"property_id": pspec.name, "non_null_object": True})
                    continue

                if parent_instance_default == instance_default:
                    continue

                retval.append(
                    {
                        "parent_owner": GObject.type_name(owner),
                        "property_id": pspec.name,
                        "new_default": self._get_default_value_from_pspec(pspec, instance_default, use_instance_default=True)
                    }
                )

        return retval

    def _get_default_value_from_pspec(self, pspec, instance_default=None, use_instance_default=False, owner=None):
        if pspec is None:
            return None

        pspec_type_name = GObject.type_name(pspec)
        default_value = instance_default if use_instance_default else pspec.get_default_value()

        if pspec.value_type == GObject.TYPE_BOOLEAN:
            return "True" if default_value != 0 else "False"
        elif GObject.type_is_a(pspec.value_type, GObject.TYPE_ENUM):
            return pspec.value_type.pytype(default_value).value_nick
        elif GObject.type_is_a(pspec.value_type, GObject.TYPE_FLAGS):
            none = pspec.value_type.pytype(0).first_value_nick
            value_nicks = pspec.value_type.pytype(default_value).value_nicks

            # Cleanup none value
            if none in value_nicks:
                value_nicks.remove(none)

            return " | ".join(value_nicks) if value_nicks else None
        elif GObject.type_is_a(pspec.value_type, GObject.TYPE_GTYPE):
            return GObject.type_name(default_value) if default_value is not None else None
        elif GObject.type_is_a(pspec.value_type, GLib.strv_get_type()):
            return "\n".join(default_value) if default_value and len(default_value) else None
        elif pspec_type_name == "GParamUnichar":
            return default_value

        return default_value

    def _cmb_types_init(self):
        if self.lib not in ["gtk+", "gtk"]:
            return

        extra_types = {}

        # Extra types for Gtk 3 and 4
        extra_types["CmbIconName"] = {"parent": "gchararray"}
        extra_types["CmbText"] = {"parent": "gchararray"}

        if self.lib == "gtk+":
            extra_types["CmbStockId"] = {"parent": "gchararray"}
        else:
            extra_types["CmbBooleanUndefined"] = {"parent": "gchararray"}
            extra_types["CmbAccessibleList"] = {"parent": "gchararray"}

        self.types.update(extra_types)

    def _gtk3_init(self):
        def get_properties(name, props):
            retval = {}

            if not props:
                return {}

            for pspec in props:
                owner = GObject.type_name(pspec.owner_type)
                type_name = GObject.type_name(pspec.value_type)
                writable = pspec.flags & GObject.ParamFlags.WRITABLE

                if owner != name or type_name.startswith("Gdk") or not writable:
                    continue

                retval[pspec.name] = {
                    "type": type_name,
                    "is_object": "GParamObject" == GObject.type_name(pspec),
                    "version": None,
                    "deprecated_version": None,
                    "construct": 1 if pspec.flags & GObject.ParamFlags.CONSTRUCT_ONLY else None,
                    "default_value": self._get_default_value_from_pspec(pspec),
                    "minimum": pspec.minimum if hasattr(pspec, "minimum") else None,
                    "maximum": pspec.maximum if hasattr(pspec, "maximum") else None,
                }

            return retval

        layout_types = {}

        # Create LayoutChild classes for GtkContainer child properties
        for name in self.types:
            if not self._type_is_a(name, "GtkContainer"):
                continue

            data = self.types[name]

            # Mark class as a container type
            data["layout"] = "container"
            props = self._container_list_child_properties(name)
            properties = get_properties(name, props)

            if len(properties) > 0:
                layout_types[f"{name}LayoutChild"] = {
                    "parent": "GObject",
                    "layout": "child",
                    "properties": properties,
                    "abstract": 1,
                }

        self.types.update(layout_types)

        if self.lib != "gtk+":
            return

        # Sum of all GtkAccessible names
        a11y_actions = set()

        # Map of which GtkAccessible is used by the class
        types_accessible = {}

        # names used in accessible
        accessible_actions = {}

        # Remove Accessible derived classes
        toremove = []
        for name in self.types:
            if self._type_is_a(name, "GtkAccessible"):
                toremove.append(name)
            elif self._type_is_a(name, "GtkWidget"):
                instance = self._get_instance_from_type(name)
                if instance:
                    accessible = instance.get_accessible()
                    if accessible:
                        accessible_type_id = GObject.type_name(accessible.__gtype__)
                        types_accessible[name] = accessible_type_id
                        actions = CmbCatalogUtils.a11y_action_get_name(accessible)
                        if actions:
                            actions = actions.split("\n")

                        if actions:
                            a11y_actions = a11y_actions.union(set(actions))
                            accessible_actions[accessible_type_id] = sorted(actions)

        for key in toremove:
            del self.types[key]

        for type_id, accessible_type in types_accessible.items():
            if accessible_type in accessible_actions:
                self.accessibility_metadata[type_id] = accessible_actions[accessible_type]

        # Accessibility support
        # Property name: (type, default value, since version)
        self.__a11y_add_ifaces_from_enum([
            (
                "Property",
                None,  # Do not check if all values are present
                {
                    "description": ["gchararray", None, None],
                    "help-text": ["gchararray", None, None],
                    "id": ["gchararray", None, None],
                    "name": ["gchararray", None, None],
                    "parent": ["GtkWidget", None, None],
                    "role": ["AtkRole", "unknown", None],
                    "table-caption-object": ["GtkWidget", None, None],
                    "table-summary": ["GtkWidget", None, None],
                    # These props give an GtkBuilder error when trying to set them
                    # "table-caption": ["gchararray", None, None], # deprecated: Unknown
                    # "table-column-description": ["gchararray", None, None], # deprecated: Unknown
                    # "table-column-header": ["GtkWidget", None, None], # deprecated: Unknown
                    # "table-row-description": ["gchararray", None, None], # deprecated: Unknown
                    # "table-row-header": ["GtkWidget", None, None], # deprecated: Unknown
                    # "value": ["gdouble", None, None], # deprecated: Unknown
                }
            ),
            (
                "Relation",
                None,  # Do not check if all values are present
                {
                    "controlled-by": ["GtkWidget", None, None],
                    "controller-for": ["GtkWidget", None, None],
                    "label-for": ["GtkWidget", None, None],
                    "labelled-by": ["GtkWidget", None, None],
                    "member-of": ["GtkWidget", None, None],
                    "node-child-of": ["GtkWidget", None, None],
                    "flows-to": ["GtkWidget", None, None],
                    "flows-from": ["GtkWidget", None, None],
                    "subwindow-of": ["GtkWidget", None, None],
                    "embeds": ["GtkWidget", None, None],
                    "embedded-by": ["GtkWidget", None, None],
                    "popup-for": ["GtkWidget", None, None],
                    "parent-window-of": ["GtkWidget", None, None],
                    "described-by": ["GtkWidget", None, None],
                    "description-for": ["GtkWidget", None, None],
                    "node-parent-of": ["GtkWidget", None, None],
                    "details": ["GtkWidget", None, None],
                    "details-for": ["GtkWidget", None, None],
                    "error-message": ["GtkWidget", None, None],
                    "error-for": ["GtkWidget", None, None],
                }
            ),
            (
                "Action",
                None,  # Do not check if all values are present
                {
                    name: ["gchararray", None, None] for name in a11y_actions
                }
            )
        ])

    def _gtk4_init(self):
        # Mark Layout classes
        for name in self.types:
            data = self.types[name]

            if self._type_is_a(name, "GtkLayoutManager"):
                data["layout"] = "manager"
            elif self._type_is_a(name, "GtkLayoutChild"):
                data["layout"] = "child"

        if self.lib != "gtk":
            return

        # Accessibility support
        # Dupe Enums that need an extra undefined value
        for enum_name, member_name, values in [
            ("Orientation", "ORIENTATION", [
                (None, "undefined", "Value is undefined"),
                (0, "horizontal", "The element is in horizontal orientation."),
                (1, "vertical", "The element is in vertical orientation.")
            ]),
            ("AccessibleTristate", "ACCESSIBLE_TRISTATE", [
                (None, "undefined", "Value is undefined"),
                (0, "false", "The state is false."),
                (1, "true", "The state is true."),
                (2, "mixed", "The state is mixed.")
            ])
        ]:
            self.enumerations[f"Cmb{enum_name}Undefined"] = {
                "parent": "enum",
                "members": {f"CMB_{member_name}_{n.upper()}": {"value": v, "nick": n, "doc": d} for (v, n, d) in values}
            }

        # Map GtkAccessibleProperty, GtkAccessibleRelation and GtkAccessibleState to Cmb prefixed types and properties
        # Property name: (type, default value, since version)
        self.__a11y_add_ifaces_from_enum([
            (
                "Property",
                "GtkAccessibleProperty",
                {
                    "autocomplete": ["GtkAccessibleAutocomplete", "none", None],
                    "description": ["gchararray", None, None],
                    "has-popup": ["gboolean", "False", None],
                    "key-shortcuts": ["gchararray", None, None],
                    "label": ["gchararray", None, None],
                    "level": ["gint64", 0, None],
                    "modal": ["gboolean", "False", None],
                    "multi-line": ["gboolean", "False", None],
                    "multi-selectable": ["gboolean", "False", None],
                    "orientation": ["CmbOrientationUndefined", "undefined", None],  # Undefined default undefined
                    "placeholder": ["gchararray", None, None],
                    "read-only": ["gboolean", "False", None],
                    "required": ["gboolean", "False", None],
                    "role-description": ["gchararray", None, None],
                    "sort": ["GtkAccessibleSort", "none", None],
                    "value-max": ["gdouble", 0, None],
                    "value-min": ["gdouble", 0, None],
                    "value-now": ["gdouble", 0, None],
                    "value-text": ["gchararray", None, None],
                    "help-text": ["gchararray", None, None],
                }
            ),
            (
                "Relation",
                "GtkAccessibleRelation",
                {
                    "active-descendant": ["GtkAccessible", None, None],
                    "col-count": ["gint64", 0, None],
                    "col-index": ["gint64", 0, None],
                    "col-index-text": ["gchararray", None, None],
                    "col-span": ["gint64", 0, None],
                    "controls": ["CmbAccessibleList", None, None],  # Reference List
                    "described-by": ["CmbAccessibleList", None, None],  # Reference List
                    "details": ["CmbAccessibleList", None, None],  # Reference List
                    "error-message": ["GtkAccessible", None, None],
                    "flow-to": ["CmbAccessibleList", None, None],  # Reference List
                    "labelled-by": ["CmbAccessibleList", None, None],  # Reference List
                    "owns": ["CmbAccessibleList", None, None],  # Reference List
                    "pos-in-set": ["gint64", 0, None],
                    "row-count": ["gint64", 0, None],
                    "row-index": ["gint64", 0, None],
                    "row-index-text": ["gchararray", None, None],
                    "row-span": ["gint64", 0, None],
                    "set-size": ["gint64", 0, None],
                }
            ),
            (
                "State",
                "GtkAccessibleState",
                {
                    "busy": ["gboolean", "False", None],
                    "checked": ["CmbAccessibleTristateUndefined", "undefined", None],
                    "disabled": ["gboolean", "False", None],
                    "expanded": ["CmbBooleanUndefined", "undefined", None],  # Undefined
                    "hidden": ["gboolean", "False", None],
                    "invalid": ["GtkAccessibleInvalidState", "false", None],
                    "pressed": ["CmbAccessibleTristateUndefined", "undefined", None],
                    "selected": ["CmbBooleanUndefined", "undefined", None],  # Undefined
                    "visited": ["CmbBooleanUndefined", "undefined", "4.12"],  # Undefined
                }
            )
        ])

        # Add GtkExpression type properties

        # To create a constant expression, use the <constant> element. If the type attribute is specified, the element content
        # is interpreted as a value of that type. Otherwise, it is assumed to be an object. For instance:
        #
        #   <constant>string_filter</constant>
        #   <constant type='gchararray'>Hello, world</constant>
        #
        self.__add_type("GtkConstantExpression", "GtkExpression", [
            ("type", "gtype", "GObject"),
            ("value", "gchararray", None),
        ])

        # To create a property expression, use the <lookup> element. It can have a type attribute to specify the object type,
        # and a name attribute to specify the property to look up.
        # The content of <lookup> can either be a string that specifies the name of the object to use, an element specifying and
        # expression to provide an object, or empty to use the this object.
        #
        #   <lookup name='search'>string_filter</lookup>
        #
        self.__add_type("GtkPropertyExpression", "GtkExpression", [
            ("type", "gtype", None),
            ("value", "gchararray", None),
            ("name", "gchararray", None),
        ])

        # To create a closure expression, use the <closure> element. The function attribute specifies what function to use for
        # the closure, and the type attribute specifies its return type.
        # The content of the element contains the expressions for the parameters. For instance:
        #
        # <closure type='gchararray' function='combine_args_somehow'>
        #   <constant type='gchararray'>File size:</constant>
        #   <lookup type='GFile' name='size'>myfile</lookup>
        # </closure>
        #
        self.__add_type("GtkClosureExpression", "GtkExpression", [
            ("type", "gtype", None),
            ("function", "gchararray", None),
        ])

    def __add_type(self, type_name, parent_type, properties):
        self.types.update({
            type_name: {
                "parent": parent_type,
                "properties": {
                    prop: {
                        "type": type_name,
                        "default_value": default,
                        "is_object": False,
                        "version": None,
                        "deprecated_version": None,
                        "construct": None,
                        "translatable": False,
                    } for (prop, type_name, default) in properties
                }
            }
        })

    def __a11y_add_ifaces_from_enum(self, accessible_ifaces):
        accessible_types = {}

        # Create a custom interface for each Accessibility enumeration
        for enumeration, check_enum, attr in accessible_ifaces:
            if check_enum:
                data = self.enumerations.get(check_enum, None)
                for member in data["members"].values():
                    nick = member["nick"]

                    if nick not in attr:
                        doc = member["doc"]
                        logging.warning(f"Missing type value for {enumeration}:{nick} {doc}")
                        continue

            # Generate a list of properties for each enumeration member
            properties = {}
            for nick, prop_meta in attr.items():
                # Ignore properties without metadata
                if prop_meta is None:
                    continue

                type_name, default_value, since_version = prop_meta

                is_object = type_name in ["GtkWidget", "GtkAccessible"]

                # Add property to list with prefix to avoid name clashes
                properties[f"cmb-a11y-{enumeration.lower()}-{nick}"] = {
                    "type": type_name,
                    "is_object": is_object,
                    "disable_inline_object": is_object,
                    "default_value": default_value,
                    "version": since_version,
                    "deprecated_version": None,
                    "construct": None,
                    "translatable": type_name == "gchararray",
                }

            # Add custom interface type
            accessible_types[f"CmbAccessible{enumeration}"] = {
                "parent": "interface",
                "properties": properties,
            }

        # Add all accessible types
        self.types.update(accessible_types)

    def _type_get_properties(self, element, props, owner=None):
        retval = {}
        pspecs = {}

        if props is not None:
            for p in props:
                pspecs[p.name] = p

        for child in element.iterfind("property", nsmap):
            if child.get("writable") != "1":
                continue

            name = child.get("name")
            type_node = child.find("type", nsmap)

            # <type> might be inside an <array>
            if type_node is None:
                type_node = child.find("array", nsmap)

            if type_node is None:
                continue

            # Property pspec
            pspec = pspecs.get(name, None)

            pspec_type_name = GObject.type_name(pspec) if pspec else None

            type_name = self.pspec_map.get(pspec_type_name, None)
            if type_name is None:
                self.ignored_pspecs.add(pspec_type_name)
                continue

            is_object = type_name == "object"
            if type_name == "object" or type_name == "enum" or type_name == "flags":
                type_name = type_node.get("name", "GObject")

                if type_name.find(".") >= 0:
                    nstype_name = self.external_types.get(type_name, None)

                    if nstype_name is None:
                        self.ignored_types.add(type_name)
                        continue

                    type_name = nstype_name
                elif type_name != "GObject":
                    type_name = self.prefix + type_name
            elif type_name == "boxed":
                type_name = GObject.type_name(pspec.value_type)

                if type_name not in self.external_nstypes:
                    self.ignored_boxed_types.add(type_name)
                    continue

            retval[name] = {
                "type": type_name,
                "is_object": is_object,
                "version": child.get("version"),
                "deprecated_version": child.get("deprecated-version"),
                "construct": child.get("construct-only"),
                "default_value": self._get_default_value_from_pspec(pspec, owner=owner),
                "minimum": pspec.minimum if hasattr(pspec, "minimum") else None,
                "maximum": pspec.maximum if hasattr(pspec, "maximum") else None,
            }

        return retval

    def _type_get_signals(self, element):
        retval = {}

        for child in element.iterfind(ns("glib", "signal")):
            name = child.get("name").replace("_", "-")
            retval[name] = {
                "version": child.get("version"),
                "deprecated_version": child.get("deprecated-version"),
                "detailed": child.get("detailed"),
            }

        return retval

    def _type_get_interfaces(self, element):
        retval = []

        for child in element.iterfind("implements", nsmap):
            name = child.get("name")
            if name.find(".") < 0:
                retval.append(self.prefix + name)
            elif name in self.external_types:
                retval.append(self.external_types[name])

        return retval

    def _list_properties(self, type_name):
        nons_name = type_name.removeprefix(self.prefix)
        klass = getattr(self.mod, nons_name)
        gtype = klass.__gtype__

        if gtype.is_a(GObject.Object) or gtype.is_interface():
            return GObject.list_properties(klass)
        else:
            return []

    def _get_type_data(self, element, name, use_instance=True, skip_types=[]):
        parent = element.get("parent")

        if parent == "Expression" and name not in ["GtkConstantExpression", "GtkPropertyExpression", "GtkClosureExpression"]:
            return None
        elif parent and parent.find(".") < 0:
            parent = self.prefix + parent
        elif parent is None:
            parent = "object"
        elif parent == "GObject.ParamSpec":
            return None
        else:
            parent = self.external_types.get(parent, "GObject")

        # Get version and deprecated-version from constructor if possible
        constructor = element.find("constructor", nsmap)
        if constructor is None:
            constructor = element

        is_container = False
        overrides = []

        if use_instance:
            instance = self._get_instance_from_type(name)
            if instance is not None:
                is_container = CmbCatalogUtils.implements_buildable_add_child(instance)
                if parent not in skip_types:
                    overrides = self._type_get_properties_overrides(name)

        props = self._list_properties(name)

        return {
            "parent": parent,
            "layout": "container" if is_container else None,
            "abstract": element.get("abstract"),
            "derivable": True if element.get(ns("glib", "type-struct")) else None,
            "version": constructor.get("version"),
            "deprecated_version": constructor.get("deprecated-version"),
            "properties": self._type_get_properties(element, props, owner=name),
            "signals": self._type_get_signals(element),
            "interfaces": self._type_get_interfaces(element),
            "overrides": overrides,
        }

    def _get_boxed_types(self, boxed_types=[]):
        retval = {}

        for name in boxed_types:
            retval[name] = {
                "parent": "boxed",
                "abstract": 0,
            }

        return retval

    def _get_types(self, namespace, types=None, skip_types=[], exclude_objects=False):
        retval = {}

        for child in namespace.iterfind("class", nsmap):
            name = child.get(ns("glib", "type-name"))

            if name is None or (exclude_objects and types is None):
                continue

            if types is None or name in types:
                data = self._get_type_data(child, name, name not in skip_types, skip_types=skip_types)
                if name and data is not None:
                    retval[name] = data

        return retval

    def _get_enumerations(self, namespace, types=None):
        retval = {}

        for child in namespace.iterfind("enumeration", nsmap):
            name = child.get(ns("glib", "type-name"))
            if name and (types is None or name in types):
                retval[name] = {"parent": "enum", "members": self._enum_flags_get_members(child)}

        return retval

    def _enum_flags_get_members(self, element):
        retval = {}

        for child in element.iterfind("member", nsmap):
            doc = child.find("doc", nsmap)
            doc_text = None

            if doc is not None:
                doc_text = " ".join(doc.text.split())

            # GLib uses the C identifier as the enum/flag name
            retval[child.get(ns("c", "identifier"))] = {
                "value": child.get("value"),
                "nick": child.get(ns("glib", "nick")),
                "doc": doc_text,
            }

        return retval

    def _get_flags(self, namespace, types=None):
        retval = {}

        for child in namespace.iterfind("bitfield", nsmap):
            name = child.get(ns("glib", "type-name"))
            if name and (types is None or name in types):
                retval[name] = {"parent": "flags", "members": self._enum_flags_get_members(child)}

        return retval

    def _get_ifaces(self, namespace, types=None, exclude_objects=False):
        retval = {}

        for child in namespace.iterfind("interface", nsmap):
            name = child.get(ns("glib", "type-name"))

            if name is None or (exclude_objects and types is None):
                continue

            if types is None or name in types:
                props = self._list_properties(name)

                retval[name] = {
                    "parent": "interface",
                    "version": child.get("version"),
                    "deprecated_version": child.get("deprecated-version"),
                    "properties": self._type_get_properties(child, props),
                    "signals": self._type_get_signals(child),
                }

        return retval

    def populate_db(self, conn):
        def major_minor_from_string(string):
            if string is None:
                return (0, 0)

            tokens = string.split(".")

            major = int(tokens[0])
            minor = int(tokens[1]) if len(tokens) > 1 else 0

            return (major, minor)

        mod_major, mod_minor = major_minor_from_string(self.version)

        def clean_ver(version):
            major, minor = major_minor_from_string(version)
            return version if major >= mod_major else None

        def db_insert_enum_flags(conn, name, data):
            parent = data.get("parent", None)
            conn.execute("INSERT INTO type (library_id, type_id, parent_id) VALUES (?, ?, ?);", (self.lib, name, parent))

            members = data.get("members", [])
            for member in members:
                m = members[member]
                conn.execute(
                    f"INSERT INTO type_{parent} (type_id, name, value, nick, doc) VALUES (?, ?, ?, ?, ?);",
                    (name, member, m["value"], m["nick"], m["doc"]),
                )

        def db_insert_iface(conn, name, data):
            parent = data.get("parent", None)
            conn.execute("INSERT INTO type (library_id, type_id, parent_id) VALUES (?, ?, ?);", (self.lib, name, parent))

        def db_insert_type(conn, name, data):
            parent = data.get("parent", None)

            if parent and parent.find(".") >= 0:
                parent = "object"

            conn.execute(
                """
                INSERT INTO type (library_id, type_id, parent_id, version, deprecated_version, abstract, derivable, layout)
                VALUES (?, ?, ?, ?, ?, ?, ?, ?);
                """,
                (
                    self.lib,
                    name,
                    parent,
                    clean_ver(data.get("version", None)),
                    clean_ver(data.get("deprecated_version", None)),
                    data.get("abstract", None),
                    data.get("derivable", None),
                    data.get("layout", None),
                ),
            )

        def db_insert_type_data(conn, name, data):
            properties = data.get("properties", [])
            for prop in properties:
                p = properties[prop]
                prop_type = p["type"]

                # Ignore unknown types (Probably GBoxed)
                if (
                    prop_type.startswith(self.name)
                    and prop_type not in self.types
                    and prop_type not in self.flags
                    and prop_type not in self.enumerations
                    and prop_type not in self.ifaces
                ):
                    continue

                conn.execute(
                    """
                    INSERT INTO property (owner_id, property_id, type_id, is_object, construct_only, default_value, minimum,
                                          maximum, version, deprecated_version, disable_inline_object, translatable)
                    VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?);
                    """,
                    (
                        name,
                        prop,
                        prop_type,
                        p["is_object"] if prop_type not in self.builder_native_object_types else False,
                        p["construct"],
                        p.get("default_value", None),
                        p.get("minimum", None),
                        p.get("maximum", None),
                        clean_ver(p["version"]),
                        clean_ver(p["deprecated_version"]),
                        p.get("disable_inline_object", None),
                        p.get("translatable", None),
                    ),
                )

            signals = data.get("signals", {})
            for signal in signals:
                s = signals[signal]
                conn.execute(
                    "INSERT INTO signal (owner_id, signal_id, version, deprecated_version, detailed) VALUES (?, ?, ?, ?, ?);",
                    (name, signal, clean_ver(s["version"]), clean_ver(s["deprecated_version"]), s["detailed"]),
                )

            for iface in data.get("interfaces", []):
                conn.execute("INSERT INTO type_iface (type_id, iface_id) VALUES (?, ?);", (name, iface))

        def db_insert_type_overrides(conn, name, data):
            overrides = data.get("overrides", [])

            for data in overrides:
                if "parent_owner" not in data:
                    continue

                parent_owner = data["parent_owner"]
                new_default = data["new_default"]
                property_id = data["property_id"]

                # Get parent property
                for table in ["property", "external_property"]:
                    row = conn.execute(
                        f"""
                        SELECT type_id, is_object, construct_only, minimum, maximum, version, deprecated_version,
                            disable_inline_object, translatable
                        FROM {table}
                        WHERE owner_id=? AND property_id=?;
                        """,
                        (parent_owner, property_id)
                    ).fetchone()

                    if row is not None:
                        break

                if row is None:
                    logging.error(f"Error trying to find {parent_owner}::{property_id} property definition")
                    break

                (type_id, is_object, construct_only, minimum, maximum, version, deprecated_version,
                 disable_inline_object, translatable) = row

                # Save new default as a new property of the class
                conn.execute(
                    """
                    INSERT INTO property
                        (owner_id, property_id, type_id, is_object, construct_only, default_value, minimum,
                         maximum, version, deprecated_version, disable_inline_object, translatable, original_owner_id)
                    VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?);
                    """,
                    (
                        name,  # new owner of overridden property
                        property_id,
                        type_id,
                        is_object,
                        construct_only,
                        new_default,  # new_default
                        minimum,
                        maximum,
                        version,
                        deprecated_version,
                        disable_inline_object,
                        translatable,
                        parent_owner
                    ),
                )

        # Import library
        conn.execute(
            "INSERT INTO library (library_id, version, namespace, prefix, shared_library) VALUES (?, ?, ?, ?, ?);",
            (self.lib, self.version, self.name, self.prefix, self.shared_library),
        )

        # Import ifaces
        for name in self.ifaces:
            db_insert_iface(conn, name, self.ifaces[name])

        # Import enumeration
        for name in self.enumerations:
            db_insert_enum_flags(conn, name, self.enumerations[name])

        # Import bitfield
        for name in self.flags:
            db_insert_enum_flags(conn, name, self.flags[name])

        # Import types in topological order
        for name in self.sorted_types:
            if name not in self.types:
                continue
            db_insert_type(conn, name, self.types[name])

        # Now insert type data (properties, signals, etc)
        for name in self.sorted_types:
            if name not in self.types:
                continue
            db_insert_type_data(conn, name, self.types[name])

        # Now insert iface data (properties, signals, etc)
        for name in self.ifaces:
            db_insert_type_data(conn, name, self.ifaces[name])

        for name in self.sorted_types:
            if name not in self.types:
                continue
            db_insert_type_overrides(conn, name, self.types[name])

        # Get versions from all types, properties and signal of this library
        versions = [(mod_major, mod_minor)]
        for row in conn.execute(
            """
            SELECT version FROM type WHERE version IS NOT NULL AND library_id=?
            UNION
            SELECT p.version FROM property AS p, type AS t
              WHERE p.version IS NOT NULL AND p.owner_id = t.type_id AND t.library_id=? AND p.original_owner_id IS NULL
            UNION
            SELECT s.version FROM signal AS s, type AS t
              WHERE s.version IS NOT NULL AND s.owner_id = t.type_id AND t.library_id=?;
            """,
            (self.lib, self.lib, self.lib),
        ):
            major, minor = major_minor_from_string(row[0])

            if major >= mod_major:
                versions.append((major, minor))

        versions = sorted(list(dict.fromkeys(versions)))

        # Save target versions
        for major, minor in versions:
            conn.execute("INSERT INTO library_version (library_id, version) VALUES (?, ?);", (self.lib, f"{major}.{minor}"))



#
# Unit Tests utils
#
# Copyright (C) 2023-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#

import io
import cairo
import struct
import time

from gi.repository import GLib, Gtk


# Based on Gtk sources gtktestutils.c gtk_test_widget_wait_for_draw()
def wait_for_drawing(window):
    done = {"done": False}
    main_loop = GLib.MainContext.default()

    def quit_main_loop_callback(widget, frame_clock, done):
        done["done"] = True
        main_loop.wakeup()
        return GLib.SOURCE_REMOVE

    window.add_tick_callback(quit_main_loop_callback, done)

    while not done["done"]:
        main_loop.iteration(True)


def surface_write_ppm(surface, path):
    data = surface.get_data()
    w = surface.get_width()
    h = surface.get_height()

    with open(path, "wb") as fd:
        fd.write(bytes(f"P6\n#Created by Cambalache tests/utils.py\n#\n{w} {h}\n255\n", "utf-8"))
        for r, g, b, a in struct.iter_unpack("BBBB", data):
            fd.write(bytes([r, g, b]))
        fd.close()


def window_screenshot(window):
    if window is None:
        return None

    # Wait for window to finish drawing
    wait_for_drawing(window)

    paintable = Gtk.WidgetPaintable.new(window)
    snapshot = Gtk.Snapshot()

    w = paintable.get_intrinsic_width()
    h = paintable.get_intrinsic_height()

    # Draw widget to cairo surface
    surface = cairo.ImageSurface(cairo.FORMAT_RGB24, w, h)

    paintable.snapshot(snapshot, w, h)
    node = snapshot.to_node()

    cr = cairo.Context(surface)
    node.draw(cr)

    surface.flush()

    return surface


def image_reload_as_png(surface):
    screenshot_png_bytes = io.BytesIO()
    surface.write_to_png(screenshot_png_bytes)
    screenshot_png_bytes.seek(0)
    return cairo.ImageSurface.create_from_png(screenshot_png_bytes)


def mean_squared_error(original, screenshot, ignore_color=None):
    original_data = original.get_data()
    screenshot_data = screenshot.get_data()

    n = original.get_width() * original.get_height()

    if len(original_data) != len(screenshot_data):
        raise ValueError("Screenshot has different size")

    # Calculate mean squared error for each channel
    red = 0.0
    green = 0.0
    blue = 0.0
    for i in range(0, n):
        offset = i * 4
        r, g, b, a = struct.unpack_from("BBBB", original_data, offset)
        rr, gg, bb, aa = struct.unpack_from("BBBB", screenshot_data, offset)

        if ignore_color is not None:
            ir, ig, ib = ignore_color
            if r == ir and g == ig and b == ib:
                continue

        red += (r / 255.0 - rr / 255.0) ** 2
        green += (g / 255.0 - gg / 255.0) ** 2
        blue += (b / 255.0 - bb / 255.0) ** 2

    total = 0.2126 * red + 0.7152 * green + 0.0722 * blue

    return red / n, green / n, blue / n, total / n


def process_all_pending_gtk_events(deciseconds=3):
    main_loop = GLib.MainContext.default()

    for i in range(deciseconds*10):
        while main_loop.pending():
            main_loop.iteration(False)
        time.sleep(0.01)


def __get_children(obj):
    if obj is None:
        return []

    retval = []

    child = obj.get_first_child()
    while child is not None:
        retval.append(child)
        child = child.get_next_sibling()
    return retval


def find_by_buildable_id(widget, name):
    retval = None

    if isinstance(widget, Gtk.Buildable) and Gtk.Buildable.get_buildable_id(widget) == name:
        return widget

    for child in __get_children(widget):
        retval = find_by_buildable_id(child, name)
        if retval:
            return retval

        if isinstance(child, Gtk.Buildable) and Gtk.Buildable.get_buildable_id(child) == name:
            return child

    return retval


def cmb_create_app():
    from cambalache.app import CmbApplication

    settings = Gtk.Settings.get_default()
    settings.props.gtk_enable_animations = False

    app = CmbApplication()
    app.register()
    app.activate()

    window = None

    # Spin until we get the main window
    main_loop = GLib.MainContext.default()
    while main_loop.pending() and not window:
        main_loop.iteration(False)

        # Get window if any
        windows = app.get_windows()
        if len(windows):
            window = windows[0]

    return app, window


def cmb_destroy_app(app, window):
    window.activate_action("close", None)
    process_all_pending_gtk_events()
    window.destroy()
    app.quit()

    process_all_pending_gtk_events()



#
# CmbToplevelChooser
#
# Copyright (C) 2021-2023  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk, Gio
from cambalache import CmbObject, CmbUI


class CmbNoneObject(CmbObject):
    __gtype_name__ = "CmbNoneObject"

    @GObject.Property(type=str)
    def display_name_type(self):
        return "(None)"

    @GObject.Property(type=str)
    def display_name(self):
        return "<b>(None)</b>"


class CmbToplevelChooser(Gtk.DropDown):
    __gtype_name__ = "CmbToplevelChooser"

    object = GObject.Property(type=CmbUI, flags=GObject.ParamFlags.READWRITE)
    derivable_only = GObject.Property(type=bool, default=False, flags=GObject.ParamFlags.READWRITE)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.none_list = Gio.ListStore(item_type=CmbNoneObject)
        self.none_list.append(CmbNoneObject())

        self.__update_model()

        self.__expression = Gtk.PropertyExpression.new(CmbObject, None, "display-name-type")
        self.props.expression = self.__expression

        self.connect("notify::object", self.__on_object_notify)
        self.connect("notify::selected-item", self.__on_selected_item_notify)

    def __update_model(self):
        if self.object is None:
            self.props.model = None
            return

        lists = Gio.ListStore(item_type=Gio.ListModel)
        lists.append(self.none_list)

        if self.derivable_only:
            lists.append(Gtk.FilterListModel(
                model=self.object,
                filter=Gtk.CustomFilter.new(lambda i, d: i.info.derivable, None)
            ))
        else:
            lists.append(self.object)

        flatten = Gtk.FlattenListModel(model=lists)
        self.props.model = flatten
        self.set_selected(0)

    def __on_object_notify(self, obj, pspec):
        self.props.model = None
        self.__update_model()

    def __on_selected_item_notify(self, obj, pspec):
        self.notify("cmb-value")

    @GObject.Property(type=int)
    def cmb_value(self):
        item = self.get_selected_item()
        if item is None:
            return 0

        return item.object_id

    @cmb_value.setter
    def _set_cmb_value(self, value):
        if self.object is None or value is None or value == 0:
            self.set_selected(0)
            return

        model = self.props.model
        item = self.object.project.get_object_by_id(self.object.ui_id, value)

        if item is not None:
            for position in range(1, model.props.n_items + 1):
                i = model.get_item(position)
                if i == item:
                    self.set_selected(position)
                    break
        else:
            self.set_selected(0)



import os
import gi

basedir = os.path.join(os.path.split(os.path.dirname(__file__))[0])

# Ensure home directory
homedir = os.path.join(basedir, ".local", "home")
os.makedirs(os.path.join(homedir, "Projects"), exist_ok=True)

os.environ["CMB_NOTIFICATION_URL"] = "https://dummybackend.local:2000"
os.environ["GSETTINGS_BACKEND"] = "memory"
os.environ["HOME"] = homedir

gi.require_version("Gtk", "4.0")
from gi.repository import Gtk # noqa E402
from tools.cmb_init_dev import cmb_init_dev # noqa E402

# Disable animations
settings = Gtk.Settings.get_default()
settings.props.gtk_enable_animations = False

# Make sure we can run Cambalache from sources
cmb_init_dev()




#
# Cambalache Property Type Info wrapper
#
# Copyright (C) 2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject
from .cmb_base_objects import CmbBasePropertyInfo

from cambalache import getLogger

logger = getLogger(__name__)


class CmbPropertyInfo(CmbBasePropertyInfo):
    internal_child = GObject.Property(type=GObject.GObject, flags=GObject.ParamFlags.READWRITE)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.is_a11y = CmbPropertyInfo.type_is_accessible(self.owner_id)
        self.a11y_property_id = CmbPropertyInfo.accessible_property_remove_prefix(self.owner_id, self.property_id)

    @classmethod
    def type_is_accessible(cls, owner_id):
        return owner_id in [
            "CmbAccessibleProperty",
            "CmbAccessibleRelation",
            "CmbAccessibleState",
            "CmbAccessibleAction"
        ]

    @classmethod
    def accessible_property_remove_prefix(cls, owner_id, property_id):
        prefix = {
                "CmbAccessibleProperty": "cmb-a11y-property-",
                "CmbAccessibleRelation": "cmb-a11y-relation-",
                "CmbAccessibleState": "cmb-a11y-state-",
                "CmbAccessibleAction": "cmb-a11y-action-"
            }.get(owner_id, None)

        if prefix is None:
            return None

        # A11y property name without prefix
        return property_id.removeprefix(prefix)




#
# Cambalache Base File Monitor
#
# Copyright (C) 2025  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

import os
from gi.repository import GObject, Gio

from .cmb_base import CmbBase
from cambalache import getLogger

logger = getLogger(__name__)


class FileStatus(GObject.GEnum):
    NONE = 0
    DELETED = 1
    CHANGED = 2
    NOT_FOUND = 3
    RENAMED = 4


# FIXME: this should be a GInterface
class CmbBaseFileMonitor(CmbBase):
    __gtype_name__ = "CmbBaseFileMonitor"

    file_status = GObject.Property(type=FileStatus, default=FileStatus.NONE, flags=GObject.ParamFlags.READWRITE)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.saving = False
        self.monitor = None
        self.gfile = None
        self.new_filename = None

    def __on_file_changed(self, file_monitor, file, other_file, event_type):
        if self.saving:
            return

            self.file_status = FileStatus.CHANGED
        elif event_type in [Gio.FileMonitorEvent.DELETED, Gio.FileMonitorEvent.MOVED_OUT]:
            # TODO: if moved out but still inside the directory hierarchy we should update the path
            self.file_status = FileStatus.DELETED
        elif event_type == Gio.FileMonitorEvent.RENAMED:
            if self.file_status == FileStatus.NOT_FOUND:
                self.file_status = FileStatus.CHANGED
                return

            filename = other_file.get_path()
            projectdir = os.path.dirname(self.project.filename) if self.project.filename else "."
            self.new_filename = os.path.relpath(filename, projectdir)
            self.file_status = FileStatus.RENAMED

    def update_file_monitor(self, filename):
        if self.monitor:
            self.monitor.cancel()

        old_path = None
        if self.gfile:
            old_path = self.gfile.get_path()
            self.gfile = None

        if not filename:
            return

        if not os.path.isabs(filename):
            if self.project.filename is None:
                return

            dirname = os.path.dirname(self.project.filename)
            fullpath = os.path.join(dirname, filename)
        else:
            fullpath = filename

        if old_path and os.path.exists(old_path) and old_path != fullpath:
            # Clear file state
            if filename in self.project._file_state:
                self.project._file_state.pop(filename)

            # Rename and force project save
            os.rename(old_path, fullpath)
            self.project.save()

        self.gfile = Gio.File.new_for_path(fullpath)
        self.monitor = self.gfile.monitor(Gio.FileMonitorFlags.WATCH_MOVES, None)
        self.monitor.connect("changed", self.__on_file_changed)

    def reload(self):
        self.file_status = FileStatus.NONE

        # Clear history
        self.project.history_enabled = True
        self.project.clear_history()



#
# DbCodegen - Cambalache DB Code Generator
#
# Copyright (C) 2021-2024  Juan Pablo Ugarte
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#

import os
import sys
import sqlite3


class CambalacheDb:
    def __init__(self):
        # Create DB file
        self.conn = sqlite3.connect(":memory:")

        dirname = os.path.dirname(__file__) or "."

        # Create DB tables
        with open(os.path.join(dirname, "../cambalache/db/cmb_base.sql"), "r") as sql:
            self.conn.executescript(sql.read())

        with open(os.path.join(dirname, "../cambalache/db/cmb_project.sql"), "r") as sql:
            self.conn.executescript(sql.read())

        self.conn.commit()

    def _get_table_data(self, table):
        c = self.conn.cursor()
        columns = []

        for row in c.execute(f"PRAGMA table_info({table});"):
            col = row[1]
            col_type = row[2]
            pk = row[5]

            if col_type == "INTEGER":
                col_type = "int"
            elif col_type == "TEXT":
                col_type = "str"
            elif col_type == "BOOLEAN":
                col_type = "bool"
            elif col_type == "REAL":
                col_type = "float"
            else:
                print("Error unknown type", col_type)

            columns.append({"name": col, "type": col_type, "pk": pk})

        c.close()

        return columns

    def dump_table_as_class(self, fd, table, klass, parent="CmbBase", mutable=False, construct_only=[]):
        c = self.conn.cursor()
        columns = self._get_table_data(table)

        fd.write(f"\n\nclass {klass}({parent}):\n")
        fd.write(f'    __gtype_name__ = "{klass}"\n\n')

        # PKs
        all_pk_columns = ""
        pks = []
        for col in columns:
            name = col['name']
            prop_type = col['type']

            if mutable and not col["pk"] and name not in construct_only:
                continue

            fd.write(f"    {name} = GObject.Property(type={prop_type}")
            fd.write(", flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT_ONLY")

            if prop_type == "bool":
                fd.write(", default=False")

            fd.write(")\n")

            if name in construct_only:
                continue

            pks.append(name)
            all_pk_columns += f"self.{name}, "

        all_columns = ""
        all_columns_assign = ""
        for col in columns:
            all_columns += ", " + col["name"]
            if mutable and not col["pk"]:
                continue
            all_columns_assign += f",\n                   {col['name']}={col['name']}"

        _pk_columns = f"({', '.join(pks)})"
        _pk_values = f"({', '.join(['?' for i in range(len(pks))])})"

        # Init
        fd.write("\n    def __init__(self, **kwargs):\n")
        fd.write("        super().__init__(**kwargs)\n")

        # Class from_row()
        fd.write("\n    @classmethod\n")
        fd.write(f"    def from_row(cls, project{all_columns}):\n")
        fd.write(f"        return cls(project=project{all_columns_assign})\n")

        if mutable:
            for col in columns:
                name = col['name']
                if col["pk"] or name in construct_only:
                    continue

                fd.write(f"\n    @GObject.Property(type={col['type']}")
                if col["type"] == "bool":
                    fd.write(", default = False")
                fd.write(")\n")
                fd.write(f"    def {name}(self):\n")
                fd.write(
                    f"        return self.db_get(\"SELECT {name} FROM {table} WHERE {_pk_columns} IS {_pk_values};\",\n"
                )
                fd.write(f"                           ({all_pk_columns}))\n")

                fd.write(f"\n    @{name}.setter\n")
                fd.write(f"    def _set_{name}(self, value):\n")
                fd.write(f"        self.db_set(\"UPDATE {table} SET {name}=? WHERE {_pk_columns} IS {_pk_values};\",\n")
                fd.write(f"                    ({all_pk_columns}), value)\n")

        c.close()

    def dump(self, filename):
        with open(filename, "w") as fd:
            fd.write(
                """# flake8: noqa
# THIS FILE IS AUTOGENERATED, DO NOT EDIT!!!
#
# Cambalache Base Object wrappers
#
# Copyright (C) 2021-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject
from .cmb_base import CmbBase
from .cmb_base_file_monitor import CmbBaseFileMonitor
"""
            )

            # Base Objects
            self.dump_table_as_class(fd, "library", "CmbBaseLibraryInfo", mutable=True)
            self.dump_table_as_class(fd, "property", "CmbBasePropertyInfo")
            self.dump_table_as_class(fd, "signal", "CmbSignalInfo")
            self.dump_table_as_class(fd, "type", "CmbBaseTypeInfo")
            self.dump_table_as_class(fd, "type_data", "CmbBaseTypeDataInfo")
            self.dump_table_as_class(fd, "type_data_arg", "CmbBaseTypeDataArgInfo")
            self.dump_table_as_class(fd, "type_child_type", "CmbTypeChildInfo")
            self.dump_table_as_class(fd, "type_internal_child", "CmbBaseTypeInternalChildInfo")

            # Project Objects
            self.dump_table_as_class(fd, "ui", "CmbBaseUI", mutable=True, parent="CmbBaseFileMonitor")
            self.dump_table_as_class(fd, "css", "CmbBaseCSS", mutable=True, parent="CmbBaseFileMonitor")
            self.dump_table_as_class(fd, "gresource", "CmbBaseGResource", mutable=True, parent="CmbBaseFileMonitor", construct_only=["resource_type"])
            self.dump_table_as_class(fd, "object_property", "CmbBaseProperty", mutable=True)
            self.dump_table_as_class(fd, "object_layout_property", "CmbBaseLayoutProperty", mutable=True)
            self.dump_table_as_class(fd, "object_signal", "CmbSignal", mutable=True)
            self.dump_table_as_class(fd, "object", "CmbBaseObject", mutable=True)
            self.dump_table_as_class(fd, "object_data", "CmbBaseObjectData", mutable=True)
            fd.close()

        os.system(f"black {filename}")


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print(f"Usage: {sys.argv[0]} output.py")
        exit()

    db = CambalacheDb()
    db.dump(sys.argv[1])



# AdwDialog Controller
#
# Copyright (C) 2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Adw, CambalachePrivate
from merengue.mrg_gtk import MrgGtkWidget, MrgSelection
from merengue import MrgPlaceholder


class MrgAdwDialog(MrgGtkWidget):
    object = GObject.Property(type=Adw.Dialog, flags=GObject.ParamFlags.READWRITE)

    def __init__(self, **kwargs):
        self.__headerbar_height = 0
        super().__init__(**kwargs)

        self.selection = MrgSelection(app=self.app, container=self.object)

    def object_changed(self, old, new):
        if old:
            old.close()

        # Handle widget selection
        if self.selection:
            self.selection.container = self.object

        self.on_selected_changed()

        if self.object is None:
            return

        # Make sure we call adw_dialog_present()
        self.object.present(None)

        self.object.set_title(GObject.type_name(self.object.__gtype__))
        CambalachePrivate.widget_set_application_id(self.object.props.parent, f"Casilda:{self.ui_id}.{self.object_id}")
        self.__update_placeholder()

    def __update_placeholder(self):
        if self.object is None:
            return

        if len(self.get_children()) == 0:
            self.add(MrgPlaceholder(visible=True, controller=self))

    def get_children(self):
        child = self.object.get_child() if self.object else None
        return [child] if child else []

    def add(self, child):
        if self.object:
            self.object.set_child(child)



#
# CmbPropertyChooser
#
# Copyright (C) 2023-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk

from ..cmb_object import CmbObject
from ..cmb_property_info import CmbPropertyInfo


class CmbPropertyChooser(Gtk.ComboBoxText):
    __gtype_name__ = "CmbPropertyChooser"

    object = GObject.Property(type=CmbObject, flags=GObject.ParamFlags.READWRITE)
    target_info = GObject.Property(type=CmbPropertyInfo, flags=GObject.ParamFlags.READWRITE)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.__populate()
        self.connect("notify", self.__on_notify)

    def __on_notify(self, obj, pspec):
        if pspec.name in ["object", "target-info"]:
            self.__populate()

    def __populate(self):
        self.remove_all()

        if self.object is None or self.target_info is None:
            return

        target_info = self.target_info
        target_type = target_info.type_id
        target_type_info = self.object.project.type_info.get(target_type, None)
        target_is_object = target_info.is_object
        target_is_iface = target_type_info.parent_id == "interface" if target_type_info else False

        for prop in sorted(self.object.properties, key=lambda p: p.property_id):
            info = prop.info

            if info.is_a11y:
                continue

            # Ignore construct only properties
            if info.construct_only:
                continue

            source_type_info = self.object.project.type_info.get(info.type_id, None)
            source_is_object = info.is_object
            source_is_iface = source_type_info.parent_id == "interface" if source_type_info else False

            if target_is_object or target_is_iface:
                # Ignore non object properties
                if not source_is_object and not source_is_iface:
                    continue

                # Ignore object properties of a different type
                if source_type_info and not source_type_info.is_a(target_info.type_id):
                    continue
            elif source_is_object or source_is_iface:
                continue

            # Enums and Flags has to be the same type
            if target_type_info and target_type_info.parent_id in ["flags", "enum"] and info.type_id != target_type:
                continue

            if source_type_info and source_type_info.parent_id in ["flags", "enum"] and info.type_id != target_type:
                continue

            compatible = info.type_id == target_type

            if not compatible:
                try:
                    gtype_id = GObject.type_from_name(info.type_id)
                    gtarget_id = GObject.type_from_name(target_type)
                    if gtype_id and gtarget_id:
                        compatible = GObject.Value.type_compatible(gtype_id, gtarget_id)
                        if not compatible:
                            compatible = GObject.Value.type_transformable(gtype_id, gtarget_id)
                except Exception as e:  # noqa F841
                    self.append(prop.property_id, prop.property_id + "*")
                    continue

            if compatible:
                self.append(prop.property_id, prop.property_id)



#
# Cambalache Layout Property wrapper
#
# Copyright (C) 2021  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject

from .cmb_base_objects import CmbBaseLayoutProperty
from .cmb_property_info import CmbPropertyInfo
from . import utils


class CmbLayoutProperty(CmbBaseLayoutProperty):
    object = GObject.Property(type=GObject.GObject, flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT_ONLY)
    info = GObject.Property(type=CmbPropertyInfo, flags=GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT_ONLY)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.version_warning = None

        owner_info = self.project.type_info.get(self.info.owner_id, None)
        self.library_id = owner_info.library_id
        self._update_version_warning()

        self.connect("notify", self.__on_notify)

    def __str__(self):
        return f"CmbLayoutProperty<{self.object.type_id} {self.info.owner_id}:{self.property_id}>"

    def __on_notify(self, obj, pspec):
        obj = self.object
        self.project._object_layout_property_changed(obj.parent, obj, self)

    @GObject.Property(type=str)
    def value(self):
        c = self.project.db.execute(
            """
            SELECT value
            FROM object_layout_property
            WHERE ui_id=? AND object_id=? AND child_id=? AND owner_id=? AND property_id=?;
            """,
            (self.ui_id, self.object_id, self.child_id, self.owner_id, self.property_id),
        )
        row = c.fetchone()
        return row[0] if row is not None else self.info.default_value

    @value.setter
    def _set_value(self, value):
        c = self.project.db.cursor()

        if value is None or value == self.info.default_value:
            c.execute(
                """
                DELETE FROM object_layout_property
                WHERE ui_id=? AND object_id=? AND child_id=? AND owner_id=? AND property_id=?;
                """,
                (self.ui_id, self.object_id, self.child_id, self.owner_id, self.property_id),
            )
            value = None
        else:
            # Do not use REPLACE INTO, to make sure both INSERT and UPDATE triggers are used
            count = self.db_get(
                """
                SELECT count(value)
                FROM object_layout_property
                WHERE ui_id=? AND object_id=? AND child_id=? AND owner_id=? AND property_id=?;
                """,
                (self.ui_id, self.object_id, self.child_id, self.owner_id, self.property_id),
            )

            if count:
                c.execute(
                    """
                    UPDATE object_layout_property
                    SET value=?
                    WHERE ui_id=? AND object_id=? AND child_id=? AND owner_id=? AND property_id=?;
                    """,
                    (value, self.ui_id, self.object_id, self.child_id, self.owner_id, self.property_id),
                )
            else:
                c.execute(
                    """
                    INSERT INTO object_layout_property (ui_id, object_id, child_id, owner_id, property_id, value)
                    VALUES (?, ?, ?, ?, ?, ?);
                    """,
                    (self.ui_id, self.object_id, self.child_id, self.owner_id, self.property_id, value),
                )

        c.close()

    def _update_version_warning(self):
        target = self.object.ui.get_target(self.library_id)
        return utils.get_version_warning(target, self.info.version, self.info.deprecated_version, self.property_id)



#
# CmbPollOptionCheck
#
# Copyright (C) 2025  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from cambalache import getLogger
from gi.repository import GLib, GObject, Gtk

logger = getLogger(__name__)


@Gtk.Template(resource_path="/ar/xjuan/Cambalache/cmb_poll_option_check.ui")
class CmbPollOptionCheck(Gtk.CheckButton):
    __gtype_name__ = "CmbPollOptionCheck"

    label = Gtk.Template.Child()
    bar = Gtk.Template.Child()

    def __init__(self, **kwargs):
        self.__fraction = None
        self.__tick_id = None
        super().__init__(**kwargs)

    @GObject.Property(type=str)
    def option(self):
        return self.label.props.label

    @option.setter
    def _set_option(self, option):
        self.label.props.label = option

    @GObject.Property(type=float)
    def fraction(self):
        return self.__fraction

    @fraction.setter
    def _set_fraction(self, fraction):
        self.__fraction = fraction

        if fraction < 0:
            self.bar.props.visible = False
        else:
            self.bar.props.visible = True
            if self.__tick_id is None:
                self.__tick_id = self.add_tick_callback(self.__update_fraction)

    def __update_fraction(self, widget, frame_clock):
        if self.bar.props.fraction < self.__fraction:
            self.bar.props.fraction = min(self.__fraction, self.bar.props.fraction + 0.08)
        elif self.bar.props.fraction > self.__fraction:
            self.bar.props.fraction = max(self.__fraction, self.bar.props.fraction - 0.08)
        else:
            self.__tick_id = None
            return GLib.SOURCE_REMOVE

        return GLib.SOURCE_CONTINUE



#
# CmbUIEditor - Cambalache UI Editor
#
# Copyright (C) 2025  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk

from .cmb_object import CmbObject


@Gtk.Template(resource_path="/ar/xjuan/Cambalache/cmb_object_editor.ui")
class CmbObjectEditor(Gtk.Box):
    __gtype_name__ = "CmbObjectEditor"

    switcher = Gtk.Template.Child()
    stack = Gtk.Template.Child()
    property_editor = Gtk.Template.Child()
    layout_editor = Gtk.Template.Child()
    signal_editor = Gtk.Template.Child()
    fragment_editor = Gtk.Template.Child()
    accessible_editor = Gtk.Template.Child()

    layout_page = Gtk.Template.Child()
    signals_page = Gtk.Template.Child()
    fragment_page = Gtk.Template.Child()
    accessible_page = Gtk.Template.Child()

    def __init__(self, **kwargs):
        self._object = None

        super().__init__(**kwargs)

    @GObject.Property(type=CmbObject)
    def object(self):
        return self._object

    @object.setter
    def _set_object(self, obj):
        self._object = obj

        is_not_builtin = not obj.info.is_builtin if obj else True
        self.property_editor.object = obj

        for editor, page in [
            (self.layout_editor, self.layout_page),
            (self.signal_editor, self.signals_page),
            (self.accessible_editor, self.accessible_page),
            (self.fragment_editor, self.fragment_page),
        ]:
            editor.object = obj
            page.props.visible = is_not_builtin

        # TODO: add metadata in catalog to disable this from data when the special cases grow too much
        if obj and obj.info.is_a("GtkEventController"):
            for page in [self.layout_page, self.accessible_page, self.fragment_page]:
                page.props.visible = False


Gtk.WidgetClass.set_css_name(CmbObjectEditor, "CmbObjectEditor")



#
# CmbTextView
#
# Copyright (C) 2021-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk
from .cmb_text_buffer import CmbTextBuffer


@Gtk.Template(resource_path="/ar/xjuan/Cambalache/control/cmb_text_view.ui")
class CmbTextView(Gtk.Box):
    __gtype_name__ = "CmbTextView"

    __gsignals__ = {
        "edit-translatable": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    cmb_value = GObject.Property(type=str, flags=GObject.ParamFlags.READWRITE)

    buffer = Gtk.Template.Child()
    view = Gtk.Template.Child()
    edit = Gtk.Template.Child()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        GObject.Object.bind_property(
            self,
            "cmb-value",
            self.buffer,
            "cmb-value",
            GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL,
        )

    @GObject.Property(type=bool, default=False)
    def translatable(self):
        return self.edit.props.visible

    @translatable.setter
    def _set_translatable(self, value):
        self.edit.props.visible = value

    @Gtk.Template.Callback("on_edit_clicked")
    def __on_edit_clicked(self, widget):
        self.emit("edit-translatable")

Gtk.WidgetClass.set_css_name(CmbTextView, "CmbTextView")



import json
from gi.repository import Gio

from . import getLogger

logger = getLogger(__name__)


class MrgCommand():
    def init_command(self, command_socket=None):

        self.__cancellable = None
        self.__gsocket = None
        self.__conn = None
        self.__data_input = None
        self.__commands = []
        self.__writing_command = False

        if command_socket:
            self.init_connection(command_socket)

    def __on_data_output(self, source, res):
        try:
            status, bytes_written = source.write_all_finish(res)
        except Exception as e:
            if e.code != Gio.IOErrorEnum.CANCELLED:
                logger.warning(f"Error writing command {e}")
            self.close_connection()
            return

        self.__writing_command = False
        self.__consume_next_command()

    def __consume_next_command(self):
        if self.__conn is None or self.__writing_command or not self.__commands:
            return

        self.__writing_command = True
        self.__conn.props.output_stream.write_all_async(
            self.__commands.pop(0),
            0,
            self.__cancellable,
            self.__on_data_output
        )

    def write_command(self, command, args=None):
        cmd = {"command": command}

        if args is not None:
            cmd["args"] = args

        cmd = json.dumps(cmd) + "\n"

        self.__commands.append(cmd.encode())
        self.__consume_next_command()

    def handle_command(self, line):
        pass

    def __on_data_input(self, source, res):
        try:
            line, len = source.read_line_finish_utf8(res)
        except Exception as e:
            if e.code != Gio.IOErrorEnum.CANCELLED:
                logger.warning(f"Error reading command {e}")
            self.close_connection()
            return

        if line is not None and line != "":
            self.handle_command(line)

        # Queue next command read
        source.read_line_async(0, self.__cancellable, self.__on_data_input)

    def init_connection(self, fd):
        self.__cancellable = Gio.Cancellable()

        self.__gsocket = Gio.Socket.new_from_fd(fd)
        self.__conn = Gio.SocketConnection(socket=self.__gsocket)
        self.__data_input = Gio.DataInputStream.new(self.__conn.props.input_stream)

        # Queue first command read
        self.__data_input.read_line_async(0, self.__cancellable, self.__on_data_input)

    def close_connection(self):
        # Cancel all pending IO
        if self.__cancellable:
            self.__cancellable.cancel()
            self.__cancellable = None

        if self.__conn:
            self.__conn.close()
            self.__conn = None
            self.__data_input = None
            self.__gsocket = None

        self.__commands = []
        self.__writing_command = False



#
# CmbProjectSettings - Cambalache Project Editor
#
# Copyright (C) 2025  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

import os

from gi.repository import GObject, Gtk, Adw, Gio
from .cmb_project import CmbProject
from .cmb_library_info import CmbLibraryInfo
from .cmb_library_row import CmbLibraryRow
from cambalache import _, getLogger


logger = getLogger(__name__)


@Gtk.Template(resource_path="/ar/xjuan/Cambalache/cmb_project_settings.ui")
class CmbProjectSettings(Adw.PreferencesDialog):
    __gtype_name__ = "CmbProjectSettings"

    catalog_group = Gtk.Template.Child()
    icontheme_group = Gtk.Template.Child()

    def __init__(self, **kwargs):
        self._project = None

        super().__init__(**kwargs)

    @GObject.Property(type=CmbProject)
    def project(self):
        return self._project

    @project.setter
    def _set_project(self, project):
        if project == self._project:
            return

        # TODO: clear groups

        self._project = project

        if project is None:
            return

        third_party_catalogs = Gio.ListStore(item_type=CmbLibraryInfo)

        for library_id, info in self.project.library_info.items():
            if info.third_party:
                third_party_catalogs.append(info)

        self.catalog_group.bind_model(third_party_catalogs, self.__catalog_create_row, None)
        self.icontheme_group.bind_model(project.icontheme_search_paths, self.__icontheme_create_row, None)

    def __catalog_create_row(self, item, data):
        return CmbLibraryRow(info=item)

    def __icontheme_create_row(self, item, data):
        box = Gtk.Box()
        label = Gtk.Label(label=item.props.string, hexpand=True, halign=Gtk.Align.START)
        button = Gtk.Button(icon_name="list-remove-symbolic", has_frame=False)
        box.append(label)
        box.append(button)
        button.connect("clicked", self.__on_icontheme_remove_button_clicked, item)

        return Adw.PreferencesRow(child=box)

    def __on_icontheme_remove_button_clicked(self, button, item):
        position = self.project.icontheme_search_paths.find(item.props.string)
        self.project.icontheme_search_paths.remove(position)

    @Gtk.Template.Callback("on_icontheme_add_button_clicked")
    def __on_icontheme_add_button_clicked(self, button):
        dialog = Gtk.FileDialog(
            modal=True,
            title=_("Select icon path search directory"),
        )

        if self.project and self.project.filename:
            dialog.set_initial_folder(Gio.File.new_for_path(os.path.dirname(self.project.filename)))

        def dialog_callback(dialog, res):
            try:
                iconpath = dialog.select_folder_finish(res).get_path()
                basename, relpath = self.project._get_basename_relpath(iconpath)
                self.project.icontheme_search_paths.append(relpath)
            except Exception as e:
                logger.warning(f"Error {e}")

        dialog.select_folder(self.props.root, None, dialog_callback)


Gtk.WidgetClass.set_css_name(CmbProjectSettings, "CmbProjectSettings")



#
# CmbTutorial
#
# Copyright (C) 2021-2024  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from .cmb_tutor import CmbTutorPosition
from cambalache import _

intro = [
    (_("Hi, I will show you around Cambalache"), "intro_button", 5),
    (_("You can open a project and find recently used"), "open_button", 5),
    (_("Common actions like Undo"), "undo_button", 4),
    (_("Redo"), "redo_button", 2),
    (_("and Add new UI are directly accessible in the headerbar"), "add_button", 3),
    (_("together with the main menu"), "menu_button", 3),
    (
        _("Where you can create a new project"),
        _("New Project"),
        5,
        None,
        CmbTutorPosition.LEFT,
    ),
    (
        _("Import UI files"),
        _("Import"),
        3,
        None,
        CmbTutorPosition.LEFT,
    ),
    (_("Create a project to continue"), "intro_button", 2, "add-project"),
    (_("Great!"), "intro_button", 2),
    (
        _("This is the project workspace, where you can see and select the widgets to edit"),
        "view",
        6,
        None,
        CmbTutorPosition.CENTER,
    ),
    (_("Project tree, with multiple UI support"), "tree_view", 4, None, CmbTutorPosition.CENTER),
    (
        _("Class selector bar"),
        "type_chooser_box",
        3,
    ),
    (_("And the object editor"), "editor_stack", 3, None, CmbTutorPosition.CENTER),
    (_("You can search all supported classes here"), "type_chooser_all", 4, "show-type-popover", CmbTutorPosition.LEFT),
    (_("or investigate what is in each group"), "type_chooser_gtk", 4, "show-type-popover-gtk", CmbTutorPosition.LEFT),
    (_("Now let's add a new UI file"), "add_button", 5, "add-ui"),
    (_("Good, now try to create a window"), "intro_button", 4, "add-window"),
    (_("Excellent!"), "intro_button", 2),
    (_("BTW, did you know you can double click on any placeholder to create widgets?"), "intro_button", 5),
    (_("Try adding a grid"), "intro_button", 3, "add-grid"),
    (_("and a button"), "intro_button", 3, "add-button"),
    (_("Quite easy! Isn't it?"), "intro_button", 3),
    (
        _("If you have any question, contact us on Matrix!"),
        _("Contact"),
        7,
        None,
        CmbTutorPosition.LEFT,
    ),
    (
        _("That is all for now.\nIf you find Cambalache useful please consider donating"),
        _("Donate"),
        7,
        "donate",
        CmbTutorPosition.LEFT,
    ),
    (_("Have a nice day!"), "intro_button", 3, "intro-end"),
]



# GtkCenterBox Controller
#
# Copyright (C) 2021  Juan Pablo Ugarte
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation;
# version 2.1 of the License.
#
# library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors:
#   Juan Pablo Ugarte <juanpablougarte@gmail.com>
#
# SPDX-License-Identifier: LGPL-2.1-only
#

from gi.repository import GObject, Gtk

from .mrg_gtk_widget import MrgGtkWidget
from merengue import MrgPlaceholder, getLogger

logger = getLogger(__name__)


class MrgGtkCenterBox(MrgGtkWidget):
    object = GObject.Property(type=Gtk.CenterBox, flags=GObject.ParamFlags.READWRITE)

    def __init__(self, **kwargs):
        self.size = None

        super().__init__(**kwargs)

    def __get_children(self):
        return [self.object.get_start_widget(), self.object.get_center_widget(), self.object.get_end_widget()]

    def add(self, child):
        if self.object is None:
            return

    def get_child_type(self, child):
        start, center, end = self.__get_children()

        if start == child:
            return "start"
        elif center == child:
            return "center"
        elif end == child:
            return "end"

    def __ensure_placeholders(self):
        if self.object is None:
            return

        start, center, end = self.__get_children()

        if start is None:
            self.object.set_start_widget(MrgPlaceholder(visible=True, controller=self))
        if center is None:
            self.object.set_center_widget(MrgPlaceholder(visible=True, controller=self))
        if end is None:
            self.object.set_end_widget(MrgPlaceholder(visible=True, controller=self))

    def object_changed(self, old, new):
        super().object_changed(old, new)
        self.__ensure_placeholders()

    def remove_child(self, child):
        if self.object is None:
            return

        start, center, end = self.__get_children()

        if start == child:
            self.object.set_start_widget(None)
        elif center == child:
            self.object.set_center_widget(None)
        elif end == child:
            self.object.set_end_widget(None)
